"""
Install Salt on an LXC Container
================================

.. versionadded:: 2014.7.0

Please read :ref:`core config documentation <config_lxc>`.
"""

import copy
import logging
import os
import pprint
import time

import salt.client
import salt.config as config
import salt.runner
import salt.utils.cloud
import salt.utils.json
from salt.exceptions import SaltCloudSystemExit

log = logging.getLogger(__name__)

__FUN_TIMEOUT = {
    "cmd.run": 60 * 60,
    "test.ping": 10,
    "lxc.info": 40,
    "lxc.list": 300,
    "lxc.templates": 100,
    "grains.items": 100,
}
__CACHED_CALLS = {}
__CACHED_FUNS = {
    "test.ping": 3 * 60,  # cache ping for 3 minutes
    "lxc.list": 2,  # cache lxc.list for 2 seconds
}


def __virtual__():
    """
    Needs no special configuration
    """
    return True


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def _get_grain_id(id_):
    if not get_configured_provider():
        return
    infos = get_configured_provider()
    return "salt.cloud.lxc.{}.{}".format(infos["target"], id_)


def _minion_opts(cfg="minion"):
    if "conf_file" in __opts__:
        default_dir = os.path.dirname(__opts__["conf_file"])
    else:
        default_dir = (__opts__["config_dir"],)
    cfg = os.environ.get("SALT_MINION_CONFIG", os.path.join(default_dir, cfg))
    opts = config.minion_config(cfg)
    return opts


def _master_opts(cfg="master"):
    if "conf_file" in __opts__:
        default_dir = os.path.dirname(__opts__["conf_file"])
    else:
        default_dir = (__opts__["config_dir"],)
    cfg = os.environ.get("SALT_MASTER_CONFIG", os.path.join(default_dir, cfg))
    opts = config.master_config(cfg)
    opts["output"] = "quiet"
    return opts


def _client():
    return salt.client.get_local_client(mopts=_master_opts())


def _runner():
    # opts = _master_opts()
    # opts['output'] = 'quiet'
    return salt.runner.RunnerClient(_master_opts())


def _salt(fun, *args, **kw):
    """Execute a salt function on a specific minion

    Special kwargs:

            salt_target
                target to exec things on
            salt_timeout
                timeout for jobs
            salt_job_poll
                poll interval to wait for job finish result
    """
    try:
        poll = kw.pop("salt_job_poll")
    except KeyError:
        poll = 0.1
    try:
        target = kw.pop("salt_target")
    except KeyError:
        target = None
    try:
        timeout = int(kw.pop("salt_timeout"))
    except (KeyError, ValueError):
        # try to has some low timeouts for very basic commands
        timeout = __FUN_TIMEOUT.get(
            fun, 900  # wait up to 15 minutes for the default timeout
        )
    try:
        kwargs = kw.pop("kwargs")
    except KeyError:
        kwargs = {}
    if not target:
        infos = get_configured_provider()
        if not infos:
            return
        target = infos["target"]
    laps = time.time()
    cache = False
    if fun in __CACHED_FUNS:
        cache = True
        laps = laps // __CACHED_FUNS[fun]
    try:
        sargs = salt.utils.json.dumps(args)
    except TypeError:
        sargs = ""
    try:
        skw = salt.utils.json.dumps(kw)
    except TypeError:
        skw = ""
    try:
        skwargs = salt.utils.json.dumps(kwargs)
    except TypeError:
        skwargs = ""
    cache_key = (laps, target, fun, sargs, skw, skwargs)
    if not cache or (cache and (cache_key not in __CACHED_CALLS)):
        with _client() as conn:
            runner = _runner()
            rkwargs = kwargs.copy()
            rkwargs["timeout"] = timeout
            rkwargs.setdefault("tgt_type", "list")
            kwargs.setdefault("tgt_type", "list")
            ping_retries = 0
            # the target(s) have environ one minute to respond
            # we call 60 ping request, this prevent us
            # from blindly send commands to unmatched minions
            ping_max_retries = 60
            ping = True
            # do not check ping... if we are pinguing
            if fun == "test.ping":
                ping_retries = ping_max_retries + 1
            # be sure that the executors are alive
            while ping_retries <= ping_max_retries:
                try:
                    if ping_retries > 0:
                        time.sleep(1)
                    pings = conn.cmd(tgt=target, timeout=10, fun="test.ping")
                    values = list(pings.values())
                    if not values:
                        ping = False
                    for v in values:
                        if v is not True:
                            ping = False
                    if not ping:
                        raise ValueError("Unreachable")
                    break
                except Exception:  # pylint: disable=broad-except
                    ping = False
                    ping_retries += 1
                    log.error("%s unreachable, retrying", target)
            if not ping:
                raise SaltCloudSystemExit("Target {} unreachable".format(target))
            jid = conn.cmd_async(tgt=target, fun=fun, arg=args, kwarg=kw, **rkwargs)
            cret = conn.cmd(
                tgt=target, fun="saltutil.find_job", arg=[jid], timeout=10, **kwargs
            )
            running = bool(cret.get(target, False))
            endto = time.time() + timeout
            while running:
                rkwargs = {
                    "tgt": target,
                    "fun": "saltutil.find_job",
                    "arg": [jid],
                    "timeout": 10,
                }
                cret = conn.cmd(**rkwargs)
                running = bool(cret.get(target, False))
                if not running:
                    break
                if running and (time.time() > endto):
                    raise Exception(
                        "Timeout {}s for {} is elapsed".format(
                            timeout, pprint.pformat(rkwargs)
                        )
                    )
                time.sleep(poll)
            # timeout for the master to return data about a specific job
            wait_for_res = float({"test.ping": "5"}.get(fun, "120"))
            while wait_for_res:
                wait_for_res -= 0.5
                cret = runner.cmd("jobs.lookup_jid", [jid, {"__kwarg__": True}])
                if target in cret:
                    ret = cret[target]
                    break
                # recent changes
                elif "data" in cret and "outputter" in cret:
                    ret = cret["data"]
                    break
                # special case, some answers may be crafted
                # to handle the unresponsivness of a specific command
                # which is also meaningful, e.g. a minion not yet provisioned
                if fun in ["test.ping"] and not wait_for_res:
                    ret = {"test.ping": False}.get(fun, False)
                time.sleep(0.5)
            try:
                if "is not available." in ret:
                    raise SaltCloudSystemExit(
                        "module/function {} is not available".format(fun)
                    )
            except SaltCloudSystemExit:  # pylint: disable=try-except-raise
                raise
            except TypeError:
                pass
            if cache:
                __CACHED_CALLS[cache_key] = ret
    elif cache and cache_key in __CACHED_CALLS:
        ret = __CACHED_CALLS[cache_key]
    return ret


def avail_images():
    return _salt("lxc.templates")


def list_nodes(conn=None, call=None):
    hide = False
    names = __opts__.get("names", [])
    profiles = __opts__.get("profiles", {})
    profile = __opts__.get("profile", __opts__.get("internal_lxc_profile", []))
    destroy_opt = __opts__.get("destroy", False)
    action = __opts__.get("action", "")
    for opt in ["full_query", "select_query", "query"]:
        if __opts__.get(opt, False):
            call = "full"
    if destroy_opt:
        call = "full"
    if action and not call:
        call = "action"
    if profile and names and not destroy_opt:
        hide = True
    if not get_configured_provider():
        return

    path = None
    if profile and profile in profiles:
        path = profiles[profile].get("path", None)
    lxclist = _salt("lxc.list", extra=True, path=path)
    nodes = {}
    for state, lxcs in lxclist.items():
        for lxcc, linfos in lxcs.items():
            info = {
                "id": lxcc,
                "name": lxcc,  # required for cloud cache
                "image": None,
                "size": linfos["size"],
                "state": state.lower(),
                "public_ips": linfos["public_ips"],
                "private_ips": linfos["private_ips"],
            }
            # in creation mode, we need to go inside the create method
            # so we hide the running vm from being seen as already installed
            # do not also mask half configured nodes which are explicitly asked
            # to be acted on, on the command line
            if (call in ["full"] or not hide) and (
                (lxcc in names and call in ["action"]) or call in ["full"]
            ):
                nodes[lxcc] = {
                    "id": lxcc,
                    "name": lxcc,  # required for cloud cache
                    "image": None,
                    "size": linfos["size"],
                    "state": state.lower(),
                    "public_ips": linfos["public_ips"],
                    "private_ips": linfos["private_ips"],
                }
            else:
                nodes[lxcc] = {"id": lxcc, "state": state.lower()}
    return nodes


def list_nodes_full(conn=None, call=None):
    if not get_configured_provider():
        return
    if not call:
        call = "action"
    return list_nodes(conn=conn, call=call)


def show_instance(name, call=None):
    """
    Show the details from the provider concerning an instance
    """

    if not get_configured_provider():
        return
    if not call:
        call = "action"
    nodes = list_nodes_full(call=call)
    __utils__["cloud.cache_node"](nodes[name], _get_active_provider_name(), __opts__)
    return nodes[name]


def list_nodes_select(call=None):
    """
    Return a list of the VMs that are on the provider, with select fields
    """
    if not call:
        call = "select"
    if not get_configured_provider():
        return
    info = ["id", "name", "image", "size", "state", "public_ips", "private_ips"]
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full(call="action"), __opts__.get("query.selection", info), call
    )


def _checkpoint(ret):
    sret = """
id: {name}
last message: {comment}""".format(
        **ret
    )
    keys = list(ret["changes"].items())
    keys.sort()
    for ch, comment in keys:
        sret += "\n    {}:\n      {}".format(ch, comment.replace("\n", "\n      "))
    if not ret["result"]:
        if "changes" in ret:
            del ret["changes"]
        raise SaltCloudSystemExit(sret)
    log.info(sret)
    return sret


def destroy(vm_, call=None):
    """Destroy a lxc container"""
    destroy_opt = __opts__.get("destroy", False)
    profiles = __opts__.get("profiles", {})
    profile = __opts__.get("profile", __opts__.get("internal_lxc_profile", []))
    path = None
    if profile and profile in profiles:
        path = profiles[profile].get("path", None)
    action = __opts__.get("action", "")
    if action != "destroy" and not destroy_opt:
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )
    if not get_configured_provider():
        return
    ret = {"comment": "{} was not found".format(vm_), "result": False}
    if _salt("lxc.info", vm_, path=path):
        __utils__["cloud.fire_event"](
            "event",
            "destroying instance",
            "salt/cloud/{}/destroying".format(vm_),
            args={"name": vm_, "instance_id": vm_},
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )
        cret = _salt("lxc.destroy", vm_, stop=True, path=path)
        ret["result"] = cret["result"]
        if ret["result"]:
            ret["comment"] = "{} was destroyed".format(vm_)
            __utils__["cloud.fire_event"](
                "event",
                "destroyed instance",
                "salt/cloud/{}/destroyed".format(vm_),
                args={"name": vm_, "instance_id": vm_},
                sock_dir=__opts__["sock_dir"],
                transport=__opts__["transport"],
            )
            if __opts__.get("update_cachedir", False) is True:
                __utils__["cloud.delete_minion_cachedir"](
                    vm_, _get_active_provider_name().split(":")[0], __opts__
                )
    return ret


def create(vm_, call=None):
    """Create an lxc Container.
    This function is idempotent and will try to either provision
    or finish the provision of an lxc container.

    NOTE: Most of the initialization code has been moved and merged
    with the lxc runner and lxc.init functions
    """
    prov = get_configured_provider(vm_)
    if not prov:
        return
    # we cant use profile as a configuration key as it conflicts
    # with salt cloud internals
    profile = vm_.get("lxc_profile", vm_.get("container_profile", None))

    event_data = vm_.copy()
    event_data["profile"] = profile

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", event_data, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    ret = {"name": vm_["name"], "changes": {}, "result": True, "comment": ""}
    if "pub_key" not in vm_ and "priv_key" not in vm_:
        log.debug("Generating minion keys for %s", vm_["name"])
        vm_["priv_key"], vm_["pub_key"] = salt.utils.cloud.gen_keys(
            salt.config.get_cloud_config_value("keysize", vm_, __opts__)
        )
    # get the minion key pair to distribute back to the container
    kwarg = copy.deepcopy(vm_)
    kwarg["host"] = prov["target"]
    kwarg["profile"] = profile

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "requesting", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    cret = _runner().cmd("lxc.cloud_init", [vm_["name"]], kwarg=kwarg)
    ret["runner_return"] = cret
    ret["result"] = cret["result"]
    if not ret["result"]:
        ret["Error"] = "Error while creating {},".format(vm_["name"])
    else:
        ret["changes"]["created"] = "created"

    # When using cloud states to manage LXC containers
    # __opts__['profile'] is not implicitly reset between operations
    # on different containers. However list_nodes will hide container
    # if profile is set in opts assuming that it have to be created.
    # But in cloud state we do want to check at first if it really
    # exists hence the need to remove profile from global opts once
    # current container is created.
    if "profile" in __opts__:
        __opts__["internal_lxc_profile"] = __opts__["profile"]
        del __opts__["profile"]

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def get_provider(name):
    data = None
    if name in __opts__["providers"]:
        data = __opts__["providers"][name]
        if "lxc" in data:
            data = data["lxc"]
        else:
            data = None
    return data


def get_configured_provider(vm_=None):
    """
    Return the contextual provider of None if no configured
    one can be found.
    """
    if vm_ is None:
        vm_ = {}
    dalias, driver = _get_active_provider_name().split(":")
    data = None
    tgt = "unknown"
    img_provider = __opts__.get("list_images", "")
    arg_providers = __opts__.get("names", [])
    matched = False
    # --list-images level
    if img_provider:
        tgt = "provider: {}".format(img_provider)
        if dalias == img_provider:
            data = get_provider(img_provider)
            matched = True
    # providers are set in configuration
    if not data and "profile" not in __opts__ and arg_providers:
        for name in arg_providers:
            tgt = "provider: {}".format(name)
            if dalias == name:
                data = get_provider(name)
            if data:
                matched = True
                break
    # -p is providen, get the uplinked provider
    elif "profile" in __opts__:
        curprof = __opts__["profile"]
        profs = __opts__["profiles"]
        tgt = "profile: {}".format(curprof)
        if (
            curprof in profs
            and profs[curprof]["provider"] == _get_active_provider_name()
        ):
            prov, cdriver = profs[curprof]["provider"].split(":")
            tgt += " provider: {}".format(prov)
            data = get_provider(prov)
            matched = True
    # fallback if we have only __active_provider_name__
    if (__opts__.get("destroy", False) and not data) or (
        not matched and _get_active_provider_name()
    ):
        data = __opts__.get("providers", {}).get(dalias, {}).get(driver, {})
    # in all cases, verify that the linked saltmaster is alive.
    if data:
        ret = _salt("test.ping", salt_target=data["target"])
        if ret:
            return data
        else:
            log.error(
                "Configured provider %s minion: %s is unreachable",
                _get_active_provider_name(),
                data["target"],
            )
    return False



import os

import salt.modules.status as status
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class StatusTestCase(TestCase, LoaderModuleMockMixin):
    """
    test modules.status functions
    """

    def setup_loader_modules(self):
        return {status: {}}

    def _set_up_test_uptime(self):
        """
        Define common mock data for status.uptime tests
        """

        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.now = 1477004312
        m.ut = 1540154.00
        m.idle = 3047777.32
        m.ret = {
            "users": 3,
            "seconds": 1540154,
            "since_t": 1475464158,
            "days": 17,
            "since_iso": "2016-10-03T03:09:18",
            "time": "19:49",
        }

        return m

    def _set_up_test_uptime_sunos(self):
        """
        Define common mock data for cmd.run_all for status.uptime on SunOS
        """

        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.ret = {
            "retcode": 0,
            "stdout": "unix:0:system_misc:boot_time    1475464158",
        }

        return m

    def test_uptime_linux(self):
        """
        Test modules.status.uptime function for Linux
        """
        m = self._set_up_test_uptime()

        with patch.multiple(
            salt.utils.platform,
            is_linux=MagicMock(return_value=True),
            is_sunos=MagicMock(return_value=False),
            is_darwin=MagicMock(return_value=False),
            is_freebsd=MagicMock(return_value=False),
            is_openbsd=MagicMock(return_value=False),
            is_netbsd=MagicMock(return_value=False),
        ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
            status.__salt__,
            {"cmd.run": MagicMock(return_value=os.linesep.join(["1", "2", "3"]))},
        ), patch(
            "time.time", MagicMock(return_value=m.now)
        ), patch(
            "os.path.exists", MagicMock(return_value=True)
        ):
            proc_uptime = salt.utils.stringutils.to_str("{} {}".format(m.ut, m.idle))

            with patch("salt.utils.files.fopen", mock_open(read_data=proc_uptime)):
                ret = status.uptime()
                self.assertDictEqual(ret, m.ret)
            with patch("os.path.exists", MagicMock(return_value=False)):
                with self.assertRaises(CommandExecutionError):
                    status.uptime()

    def test_uptime_sunos(self):
        """
        Test modules.status.uptime function for SunOS
        """
        m = self._set_up_test_uptime()
        m2 = self._set_up_test_uptime_sunos()
        with patch.multiple(
            salt.utils.platform,
            is_linux=MagicMock(return_value=False),
            is_sunos=MagicMock(return_value=True),
            is_darwin=MagicMock(return_value=False),
            is_freebsd=MagicMock(return_value=False),
            is_openbsd=MagicMock(return_value=False),
            is_netbsd=MagicMock(return_value=False),
        ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
            status.__salt__,
            {
                "cmd.run": MagicMock(return_value=os.linesep.join(["1", "2", "3"])),
                "cmd.run_all": MagicMock(return_value=m2.ret),
            },
        ), patch(
            "time.time", MagicMock(return_value=m.now)
        ):
            ret = status.uptime()
            self.assertDictEqual(ret, m.ret)

    def test_uptime_macos(self):
        """
        Test modules.status.uptime function for macOS
        """
        m = self._set_up_test_uptime()

        kern_boottime = (
            "{{ sec = {0}, usec = {1:0<6} }} Mon Oct 03 03:09:18.23 2016".format(
                *str(m.now - m.ut).split(".")
            )
        )
        with patch.multiple(
            salt.utils.platform,
            is_linux=MagicMock(return_value=False),
            is_sunos=MagicMock(return_value=False),
            is_darwin=MagicMock(return_value=True),
            is_freebsd=MagicMock(return_value=False),
            is_openbsd=MagicMock(return_value=False),
            is_netbsd=MagicMock(return_value=False),
        ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
            status.__salt__,
            {
                "cmd.run": MagicMock(return_value=os.linesep.join(["1", "2", "3"])),
                "sysctl.get": MagicMock(return_value=kern_boottime),
            },
        ), patch(
            "time.time", MagicMock(return_value=m.now)
        ):

            ret = status.uptime()
            self.assertDictEqual(ret, m.ret)

            with patch.dict(
                status.__salt__, {"sysctl.get": MagicMock(return_value="")}
            ):
                with self.assertRaises(CommandExecutionError):
                    status.uptime()

    def test_uptime_return_success_not_supported(self):
        """
        Test modules.status.uptime function for other platforms
        """
        with patch.multiple(
            salt.utils.platform,
            is_linux=MagicMock(return_value=False),
            is_sunos=MagicMock(return_value=False),
            is_darwin=MagicMock(return_value=False),
            is_freebsd=MagicMock(return_value=False),
            is_openbsd=MagicMock(return_value=False),
            is_netbsd=MagicMock(return_value=False),
        ):
            exc_mock = MagicMock(side_effect=CommandExecutionError)
            with self.assertRaises(CommandExecutionError):
                with patch.dict(status.__salt__, {"cmd.run": exc_mock}):
                    status.uptime()

    def _set_up_test_cpustats_openbsd(self):
        """
        Define mock data for status.cpustats on OpenBSD
        """

        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.ret = {
            "0": {
                "User": "0.0%",
                "Nice": "0.0%",
                "System": "4.5%",
                "Interrupt": "0.5%",
                "Idle": "95.0%",
            }
        }

        return m

    def test_cpustats_openbsd(self):
        """
        Test modules.status.cpustats function for OpenBSD
        """
        m = self._set_up_test_cpustats_openbsd()

        systat = (
            "\n\n   1 users Load 0.20 0.07 0.05                        salt.localdomain"
            " 09:42:42\nCPU                User           Nice        System    "
            " Interrupt          Idle\n0                  0.0%           0.0%         "
            " 4.5%          0.5%         95.0%\n"
        )

        with patch.multiple(
            salt.utils.platform,
            is_linux=MagicMock(return_value=False),
            is_sunos=MagicMock(return_value=False),
            is_darwin=MagicMock(return_value=False),
            is_freebsd=MagicMock(return_value=False),
            is_openbsd=MagicMock(return_value=True),
            is_netbsd=MagicMock(return_value=False),
        ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
            status.__grains__, {"kernel": "OpenBSD"}
        ), patch.dict(
            status.__salt__, {"cmd.run": MagicMock(return_value=systat)}
        ):
            ret = status.cpustats()
            self.assertDictEqual(ret, m.ret)

    def _set_up_test_cpuinfo_bsd(self):
        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.ret = {
            "hw.model": "Intel(R) Core(TM) i5-7287U CPU @ 3.30GHz",
            "hw.ncpu": "4",
        }

        return m

    def test_cpuinfo_freebsd(self):
        m = self._set_up_test_cpuinfo_bsd()
        sysctl = "hw.model:Intel(R) Core(TM) i5-7287U CPU @ 3.30GHz\nhw.ncpu:4"

        with patch.dict(status.__grains__, {"kernel": "FreeBSD"}):
            with patch.dict(
                status.__salt__, {"cmd.run": MagicMock(return_value=sysctl)}
            ):
                ret = status.cpuinfo()
                self.assertDictEqual(ret, m.ret)

    def test_cpuinfo_openbsd(self):
        m = self._set_up_test_cpuinfo_bsd()
        sysctl = "hw.model=Intel(R) Core(TM) i5-7287U CPU @ 3.30GHz\nhw.ncpu=4"

        for bsd in ["NetBSD", "OpenBSD"]:
            with patch.dict(status.__grains__, {"kernel": bsd}):
                with patch.dict(
                    status.__salt__, {"cmd.run": MagicMock(return_value=sysctl)}
                ):
                    ret = status.cpuinfo()
                    self.assertDictEqual(ret, m.ret)

    def _set_up_test_meminfo_openbsd(self):
        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.ret = {
            "active virtual pages": "355M",
            "free list size": "305M",
            "page faults": "845",
            "pages reclaimed": "1",
            "pages paged in": "2",
            "pages paged out": "3",
            "pages freed": "4",
            "pages scanned": "5",
        }

        return m

    def test_meminfo_openbsd(self):
        m = self._set_up_test_meminfo_openbsd()
        vmstat = (
            " procs    memory       page                    disks    traps         "
            " cpu\n r   s   avm     fre  flt  re  pi  po  fr  sr cd0 sd0  int   sys  "
            " cs us sy id\n 2 103  355M    305M  845   1   2   3   4   5   0   1   21  "
            " 682   86  1  1 98"
        )

        with patch.dict(status.__grains__, {"kernel": "OpenBSD"}):
            with patch.dict(
                status.__salt__, {"cmd.run": MagicMock(return_value=vmstat)}
            ):
                ret = status.meminfo()
                self.assertDictEqual(ret, m.ret)

    def _set_up_test_w_linux(self):
        """
        Define mock data for status.w on Linux
        """

        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.ret = [
            {
                "idle": "0s",
                "jcpu": "0.24s",
                "login": "13:42",
                "pcpu": "0.16s",
                "tty": "pts/1",
                "user": "root",
                "what": "nmap -sV 10.2.2.2",
            }
        ]

        return m

    def _set_up_test_w_bsd(self):
        """
        Define mock data for status.w on Linux
        """

        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.ret = [
            {
                "idle": "0",
                "from": "10.2.2.1",
                "login": "1:42PM",
                "tty": "p1",
                "user": "root",
                "what": "nmap -sV 10.2.2.2",
            }
        ]

        return m

    def test_w_linux(self):
        m = self._set_up_test_w_linux()
        w_output = "root   pts/1  13:42    0s  0.24s  0.16s nmap -sV 10.2.2.2"

        with patch.dict(status.__grains__, {"kernel": "Linux"}):
            with patch.dict(
                status.__salt__, {"cmd.run": MagicMock(return_value=w_output)}
            ):
                ret = status.w()
                self.assertListEqual(ret, m.ret)

    def test_w_bsd(self):
        m = self._set_up_test_w_bsd()
        w_output = "root   p1 10.2.2.1    1:42PM  0 nmap -sV 10.2.2.2"

        for bsd in ["Darwin", "FreeBSD", "OpenBSD"]:
            with patch.dict(status.__grains__, {"kernel": bsd}):
                with patch.dict(
                    status.__salt__, {"cmd.run": MagicMock(return_value=w_output)}
                ):
                    ret = status.w()
                    self.assertListEqual(ret, m.ret)

    def _set_up_test_status_pid_linux(self):
        class MockData:
            """
            Store mock data
            """

        m = MockData()
        m.ret = "2701\n7539\n7540\n7542\n7623"
        return m

    def test_status_pid_linux(self):
        m = self._set_up_test_status_pid_linux()
        ps = (
            "UID      PID PPID  C STIME TTY      TIME CMD\nroot     360    2  0 Jun08 ?"
            "    00:00:00   [jbd2/dm-0-8]\nroot     947    2  0 Jun08 ?    00:00:00  "
            " [jbd2/dm-1-8]\nroot     949    2  0 Jun08 ?    00:00:09  "
            " [jbd2/dm-3-8]\nroot     951    2  0 Jun08 ?    00:00:00  "
            " [jbd2/dm-4-8]\nroot    2701    1  0 Jun08 ?    00:00:28   /usr/sbin/httpd"
            " -k start\napache  7539 2701  0 04:40 ?    00:00:04     /usr/sbin/httpd -k"
            " start\napache  7540 2701  0 04:40 ?    00:00:02     /usr/sbin/httpd -k"
            " start\napache  7542 2701  0 04:40 ?    00:01:46     /usr/sbin/httpd -k"
            " start\napache  7623 2701  0 04:40 ?    00:02:41     /usr/sbin/httpd -k"
            " start\nroot    1564    1  0 Jun11 ?    00:07:19   /usr/bin/python3"
            " /usr/bin/salt-minion -d\nroot    6674 1564  0 19:53 ?    00:00:00    "
            " /usr/bin/python3 /usr/bin/salt-call status.pid httpd -l debug"
        )

        with patch.dict(status.__grains__, {"ps": "ps -efHww"}):
            with patch.dict(
                status.__salt__, {"cmd.run_stdout": MagicMock(return_value=ps)}
            ):
                with patch.object(os, "getpid", return_value="6674"):
                    ret = status.pid("httpd")
                    self.assertEqual(ret, m.ret)



"""
Manage X509 Certificates

.. versionadded:: 2015.8.0

:depends: M2Crypto

This module can enable managing a complete PKI infrastructure including creating private keys, CAs,
certificates and CRLs. It includes the ability to generate a private key on a server, and have the
corresponding public key sent to a remote CA to create a CA signed certificate. This can be done in
a secure manner, where private keys are always generated locally and never moved across the network.

Here is a simple example scenario. In this example ``ca`` is the ca server,
and ``www`` is a web server that needs a certificate signed by ``ca``.

For remote signing, peers must be permitted to remotely call the
:mod:`sign_remote_certificate <salt.modules.x509.sign_remote_certificate>` function.


/etc/salt/master.d/peer.conf

.. code-block:: yaml

    peer:
      .*:
        - x509.sign_remote_certificate


/srv/salt/top.sls

.. code-block:: yaml

    base:
      '*':
        - cert
      'ca':
        - ca
      'www':
        - www


This state creates the CA key, certificate and signing policy. It also publishes the certificate to
the mine where it can be easily retrieved by other minions.

/srv/salt/ca.sls

.. code-block:: yaml

    /etc/salt/minion.d/x509.conf:
      file.managed:
        - source: salt://x509.conf

    restart-salt-minion:
      cmd.run:
        - name: 'salt-call service.restart salt-minion'
        - bg: True
        - onchanges:
          - file: /etc/salt/minion.d/x509.conf

    /etc/pki:
      file.directory

    /etc/pki/issued_certs:
      file.directory

    /etc/pki/ca.key:
      x509.private_key_managed:
        - bits: 4096
        - backup: True

    /etc/pki/ca.crt:
      x509.certificate_managed:
        - signing_private_key: /etc/pki/ca.key
        - CN: ca.example.com
        - C: US
        - ST: Utah
        - L: Salt Lake City
        - basicConstraints: "critical CA:true"
        - keyUsage: "critical cRLSign, keyCertSign"
        - subjectKeyIdentifier: hash
        - authorityKeyIdentifier: keyid,issuer:always
        - days_valid: 3650
        - days_remaining: 0
        - backup: True
        - require:
          - file: /etc/pki


The signing policy defines properties that override any property requested or included in a CRL. It also
can define a restricted list of minions which are allowed to remotely invoke this signing policy.

/srv/salt/x509.conf

.. code-block:: yaml

    mine_functions:
      x509.get_pem_entries: [/etc/pki/ca.crt]

    x509_signing_policies:
      www:
        - minions: 'www'
        - signing_private_key: /etc/pki/ca.key
        - signing_cert: /etc/pki/ca.crt
        - C: US
        - ST: Utah
        - L: Salt Lake City
        - basicConstraints: "critical CA:false"
        - keyUsage: "critical keyEncipherment"
        - subjectKeyIdentifier: hash
        - authorityKeyIdentifier: keyid,issuer:always
        - days_valid: 90
        - copypath: /etc/pki/issued_certs/


This state will instruct all minions to trust certificates signed by our new CA.
Using Jinja to strip newlines from the text avoids dealing with newlines in the rendered YAML,
and the  :mod:`sign_remote_certificate <salt.states.x509.sign_remote_certificate>` state will
handle properly formatting the text before writing the output.

/srv/salt/cert.sls

.. code-block:: jinja

    /usr/local/share/ca-certificates:
      file.directory

    /usr/local/share/ca-certificates/intca.crt:
      x509.pem_managed:
        - text: {{ salt['mine.get']('ca', 'x509.get_pem_entries')['ca']['/etc/pki/ca.crt']|replace('\\n', '') }}


This state creates a private key then requests a certificate signed by ca according to the www policy.

/srv/salt/www.sls

.. code-block:: yaml

    /etc/pki/www.crt:
      x509.private_key_managed:
        - name: /etc/pki/www.key
        - bits: 4096
        - backup: True

    /etc/pki/www.crt:
      x509.certificate_managed:
        - ca_server: ca
        - signing_policy: www
        - public_key: /etc/pki/www.key
        - CN: www.example.com
        - days_remaining: 30
        - backup: True

This other state creates a private key then requests a certificate signed by ca
according to the www policy but adds a strict date range for the certificate to
be considered valid.

/srv/salt/www-time-limited.sls

.. code-block:: yaml

    /etc/pki/www-time-limited.crt:
      x509.certificate_managed:
        - ca_server: ca
        - signing_policy: www
        - public_key: /etc/pki/www-time-limited.key
        - CN: www.example.com
        - not_before: 2019-05-05 00:00:00
        - not_after: 2020-05-05 14:30:00
        - backup: True

"""

import copy
import datetime
import logging
import os
import re

import salt.exceptions

try:
    from M2Crypto.RSA import RSAError
except ImportError:
    pass

log = logging.getLogger(__name__)


def __virtual__():
    """
    only load this module if the corresponding execution module is loaded
    """
    if "x509.get_pem_entry" in __salt__:
        return "x509"
    else:
        return (False, "Could not load x509 state: m2crypto unavailable")


def _revoked_to_list(revs):
    """
    Turn the mess of OrderedDicts and Lists into a list of dicts for
    use in the CRL module.
    """
    list_ = []

    for rev in revs:
        for props in rev.values():
            dict_ = {}
            for prop in props:
                for propname, val in prop.items():
                    if isinstance(val, datetime.datetime):
                        val = val.strftime("%Y-%m-%d %H:%M:%S")
                    dict_[propname] = val
            list_.append(dict_)

    return list_


def _get_file_args(name, **kwargs):
    valid_file_args = [
        "user",
        "group",
        "mode",
        "makedirs",
        "dir_mode",
        "backup",
        "create",
        "follow_symlinks",
        "check_cmd",
    ]
    file_args = {}
    extra_args = {}
    for k, v in kwargs.items():
        if k in valid_file_args:
            file_args[k] = v
        else:
            extra_args[k] = v
    file_args["name"] = name
    return file_args, extra_args


def _check_private_key(name, bits=2048, passphrase=None, new=False, overwrite=False):
    current_bits = 0
    if os.path.isfile(name):
        try:
            current_bits = __salt__["x509.get_private_key_size"](
                private_key=name, passphrase=passphrase
            )
        except salt.exceptions.SaltInvocationError:
            pass
        except RSAError:
            if not overwrite:
                raise salt.exceptions.CommandExecutionError(
                    "The provided passphrase cannot decrypt the private key."
                )

    return current_bits == bits and not new


def private_key_managed(
    name,
    bits=2048,
    passphrase=None,
    cipher="aes_128_cbc",
    new=False,
    overwrite=False,
    verbose=True,
    **kwargs
):
    """
    Manage a private key's existence.

    name:
        Path to the private key

    bits:
        Key length in bits. Default 2048.

    passphrase:
        Passphrase for encrypting the private key.

    cipher:
        Cipher for encrypting the private key.

    new:
        Always create a new key. Defaults to ``False``.
        Combining new with :mod:`prereq <salt.states.requsities.preqreq>`
        can allow key rotation whenever a new certificate is generated.

    overwrite:
        Overwrite an existing private key if the provided passphrase cannot decrypt it.

    verbose:
        Provide visual feedback on stdout, dots while key is generated.
        Default is True.

        .. versionadded:: 2016.11.0

    kwargs:
        Any kwargs supported by file.managed are supported.

    Example:

    The JINJA templating in this example ensures a private key is generated if the file doesn't exist
    and that a new private key is generated whenever the certificate that uses it is to be renewed.

    .. code-block:: jinja

        /etc/pki/www.key:
          x509.private_key_managed:
            - bits: 4096
            - new: True
            {% if salt['file.file_exists']('/etc/pki/www.key') -%}
            - prereq:
              - x509: /etc/pki/www.crt
            {%- endif %}
    """
    file_args, kwargs = _get_file_args(name, **kwargs)
    new_key = False
    if _check_private_key(
        name, bits=bits, passphrase=passphrase, new=new, overwrite=overwrite
    ):
        file_args["contents"] = __salt__["x509.get_pem_entry"](
            name, pem_type="(?:RSA )?PRIVATE KEY"
        )
    else:
        new_key = True
        file_args["contents"] = __salt__["x509.create_private_key"](
            text=True, bits=bits, passphrase=passphrase, cipher=cipher, verbose=verbose
        )

    # Ensure the key contents are a string before passing it along
    file_args["contents"] = salt.utils.stringutils.to_str(file_args["contents"])

    ret = __states__["file.managed"](**file_args)
    if ret["changes"] and new_key:
        ret["changes"] = {"new": "New private key generated"}

    return ret


def csr_managed(name, **kwargs):
    """
    Manage a Certificate Signing Request

    name:
        Path to the CSR

    properties:
        The properties to be added to the certificate request, including items like subject, extensions
        and public key. See above for valid properties.

    kwargs:
        Any arguments supported by :py:func:`file.managed <salt.states.file.managed>` are supported.

    Example:

    .. code-block:: yaml

        /etc/pki/mycert.csr:
          x509.csr_managed:
             - private_key: /etc/pki/mycert.key
             - CN: www.example.com
             - C: US
             - ST: Utah
             - L: Salt Lake City
             - keyUsage: 'critical dataEncipherment'
    """
    try:
        old = __salt__["x509.read_csr"](name)
    except salt.exceptions.SaltInvocationError:
        old = "{} is not a valid csr.".format(name)

    file_args, kwargs = _get_file_args(name, **kwargs)
    file_args["contents"] = __salt__["x509.create_csr"](text=True, **kwargs)

    ret = __states__["file.managed"](**file_args)
    if ret["changes"]:
        new = __salt__["x509.read_csr"](file_args["contents"])
        if old != new:
            ret["changes"] = {"Old": old, "New": new}

    return ret


def _certificate_info_matches(cert_info, required_cert_info, check_serial=False):
    """
    Return true if the provided certificate information matches the
    required certificate information, i.e. it has the required common
    name, subject alt name, organization, etc.

    cert_info should be a dict as returned by x509.read_certificate.
    required_cert_info should be a dict as returned by x509.create_certificate with testrun=True.
    """
    # don't modify the incoming dicts
    cert_info = copy.deepcopy(cert_info)
    required_cert_info = copy.deepcopy(required_cert_info)

    ignored_keys = [
        "Not Before",
        "Not After",
        "SHA1 Finger Print",
        "SHA-256 Finger Print",
        # The integrity of the issuer is checked elsewhere
        "Issuer Public Key",
    ]
    if __opts__["fips_mode"] is False:
        ignored_keys.append("MD5 Finger Print")
    for key in ignored_keys:
        cert_info.pop(key, None)
        required_cert_info.pop(key, None)

    if not check_serial:
        cert_info.pop("Serial Number", None)
        required_cert_info.pop("Serial Number", None)
        try:
            cert_info["X509v3 Extensions"]["authorityKeyIdentifier"] = re.sub(
                r"serial:([0-9A-F]{2}:)*[0-9A-F]{2}",
                "serial:--",
                cert_info["X509v3 Extensions"]["authorityKeyIdentifier"],
            )
            required_cert_info["X509v3 Extensions"]["authorityKeyIdentifier"] = re.sub(
                r"serial:([0-9A-F]{2}:)*[0-9A-F]{2}",
                "serial:--",
                required_cert_info["X509v3 Extensions"]["authorityKeyIdentifier"],
            )
        except KeyError:
            pass

    diff = []
    for k, v in required_cert_info.items():
        # cert info comes as byte string
        if isinstance(v, str):
            v = salt.utils.stringutils.to_bytes(v)
        try:
            if v != cert_info[k]:
                if k == "Subject Hash":
                    # If we failed the subject hash check but the subject matches, then this is
                    # likely a certificate generated under Python 2 where sorting differs and thus
                    # the hash also differs
                    if required_cert_info["Subject"] != cert_info["Subject"]:
                        diff.append(k)
                elif k == "Issuer Hash":
                    # If we failed the issuer hash check but the issuer matches, then this is
                    # likely a certificate generated under Python 2 where sorting differs and thus
                    # the hash also differs
                    if required_cert_info["Issuer"] != cert_info["Issuer"]:
                        diff.append(k)
                elif k == "X509v3 Extensions":
                    v_ext = v.copy()
                    cert_info_ext = cert_info[k].copy()
                    # DirName depends on ordering which was different on certificates created
                    # under Python 2. Remove that from the comparisson
                    try:
                        v_ext["authorityKeyIdentifier"] = re.sub(
                            r"DirName:([^\n]+)",
                            "Dirname:--",
                            v_ext["authorityKeyIdentifier"],
                        )
                        cert_info_ext["authorityKeyIdentifier"] = re.sub(
                            r"DirName:([^\n]+)",
                            "Dirname:--",
                            cert_info_ext["authorityKeyIdentifier"],
                        )
                    except KeyError:
                        pass
                    if v_ext != cert_info_ext:
                        diff.append(k)
                else:
                    diff.append(k)
        except KeyError:
            diff.append(k)

    return len(diff) == 0, diff


def _certificate_days_remaining(cert_info):
    """
    Get the days remaining on a certificate, defaulting to 0 if an error occurs.
    """
    try:
        expiry = cert_info["Not After"]
        return (
            datetime.datetime.strptime(expiry, "%Y-%m-%d %H:%M:%S")
            - datetime.datetime.now()
        ).days
    except KeyError:
        return 0


def _certificate_is_valid(name, days_remaining, append_certs, **cert_spec):
    """
    Return True if the given certificate file exists, is a certificate, matches the given specification,
    and has the required days remaining.

    If False, also provide a message explaining why.
    """
    if not os.path.isfile(name):
        return False, "{} does not exist".format(name), {}

    try:
        cert_info = __salt__["x509.read_certificate"](certificate=name)
        required_cert_info = __salt__["x509.create_certificate"](
            testrun=True, **cert_spec
        )
        if not isinstance(required_cert_info, dict):
            raise salt.exceptions.SaltInvocationError(
                "Unable to create new certificate: x509 module error: {}".format(
                    required_cert_info
                )
            )

        try:
            issuer_public_key = required_cert_info["Issuer Public Key"]
            # Verify the certificate has been signed by the ca_server or private_signing_key
            if not __salt__["x509.verify_signature"](name, issuer_public_key):
                errmsg = (
                    "Certificate is not signed by private_signing_key"
                    if "signing_private_key" in cert_spec
                    else "Certificate is not signed by the requested issuer"
                )
                return False, errmsg, cert_info
        except KeyError:
            return (
                False,
                "New certificate does not include signing information",
                cert_info,
            )

        matches, diff = _certificate_info_matches(
            cert_info, required_cert_info, check_serial="serial_number" in cert_spec
        )
        if not matches:
            return (
                False,
                "Certificate properties are different: {}".format(", ".join(diff)),
                cert_info,
            )

        actual_days_remaining = _certificate_days_remaining(cert_info)
        if days_remaining != 0 and actual_days_remaining < days_remaining:
            return (
                False,
                "Certificate needs renewal: {} days remaining but it needs to be at"
                " least {}".format(actual_days_remaining, days_remaining),
                cert_info,
            )

        return True, "", cert_info
    except salt.exceptions.SaltInvocationError as e:
        return False, "{} is not a valid certificate: {}".format(name, str(e)), {}


def _certificate_file_managed(ret, file_args):
    """
    Run file.managed and merge the result with an existing return dict.
    The overall True/False result will be the result of the file.managed call.
    """
    file_ret = __states__["file.managed"](**file_args)

    ret["result"] = file_ret["result"]
    if ret["result"]:
        ret["comment"] = "Certificate {} is valid and up to date".format(ret["name"])
    else:
        ret["comment"] = file_ret["comment"]

    if file_ret["changes"]:
        ret["changes"] = {"File": file_ret["changes"]}

    return ret


def certificate_managed(name, days_remaining=90, append_certs=None, **kwargs):
    """
    Manage a Certificate

    name
        Path to the certificate

    days_remaining : 90
        Recreate the certificate if the number of days remaining on it
        are less than this number. The value should be less than
        ``days_valid``, otherwise the certificate will be recreated
        every time the state is run. A value of 0 disables automatic
        renewal.

    append_certs:
        A list of certificates to be appended to the managed file.
        They must be valid PEM files, otherwise an error will be thrown.

    kwargs:
        Any arguments supported by :py:func:`x509.create_certificate
        <salt.modules.x509.create_certificate>` or :py:func:`file.managed
        <salt.states.file.managed>` are supported.

    not_before:
        Initial validity date for the certificate. This date must be specified
        in the format '%Y-%m-%d %H:%M:%S'.

        .. versionadded:: 3001
    not_after:
        Final validity date for the certificate. This date must be specified in
        the format '%Y-%m-%d %H:%M:%S'.

        .. versionadded:: 3001

    Examples:

    .. code-block:: yaml

        /etc/pki/ca.crt:
          x509.certificate_managed:
            - signing_private_key: /etc/pki/ca.key
            - CN: ca.example.com
            - C: US
            - ST: Utah
            - L: Salt Lake City
            - basicConstraints: "critical CA:true"
            - keyUsage: "critical cRLSign, keyCertSign"
            - subjectKeyIdentifier: hash
            - authorityKeyIdentifier: keyid,issuer:always
            - days_valid: 3650
            - days_remaining: 0
            - backup: True


    .. code-block:: yaml

        /etc/ssl/www.crt:
          x509.certificate_managed:
            - ca_server: pki
            - signing_policy: www
            - public_key: /etc/ssl/www.key
            - CN: www.example.com
            - days_valid: 90
            - days_remaining: 30
            - backup: True

    """
    if "path" in kwargs:
        name = kwargs.pop("path")

    if "ca_server" in kwargs and "signing_policy" not in kwargs:
        raise salt.exceptions.SaltInvocationError(
            "signing_policy must be specified if ca_server is."
        )

    if (
        "public_key" not in kwargs
        and "signing_private_key" not in kwargs
        and "csr" not in kwargs
    ):
        raise salt.exceptions.SaltInvocationError(
            "public_key, signing_private_key, or csr must be specified."
        )

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    is_valid, invalid_reason, current_cert_info = _certificate_is_valid(
        name, days_remaining, append_certs, **kwargs
    )

    if is_valid:
        file_args, extra_args = _get_file_args(name, **kwargs)

        return _certificate_file_managed(ret, file_args)

    if __opts__["test"]:
        file_args, extra_args = _get_file_args(name, **kwargs)
        # Use empty contents for file.managed in test mode.
        # We don't want generate a new certificate, even in memory,
        # for security reasons.
        # Using an empty string instead of omitting it will at least
        # show the old certificate in the diff.
        file_args["contents"] = ""

        ret = _certificate_file_managed(ret, file_args)

        ret["result"] = None
        ret["comment"] = "Certificate {} will be created".format(name)
        ret["changes"]["Status"] = {
            "Old": invalid_reason,
            "New": "Certificate will be valid and up to date",
        }
        return ret

    contents = __salt__["x509.create_certificate"](text=True, **kwargs)
    # Check the module actually returned a cert and not an error message as a string
    try:
        __salt__["x509.read_certificate"](contents)
    except salt.exceptions.SaltInvocationError as e:
        ret["result"] = False
        ret["comment"] = (
            "An error occurred creating the certificate {}. The result returned from"
            " x509.create_certificate is not a valid PEM file:\n{}".format(name, str(e))
        )
        return ret

    if not append_certs:
        append_certs = []
    for append_file in append_certs:
        try:
            append_file_contents = __salt__["x509.get_pem_entry"](
                append_file, pem_type="CERTIFICATE"
            )
            contents += append_file_contents
        except salt.exceptions.SaltInvocationError as e:
            ret["result"] = False
            ret["comment"] = (
                "{} is not a valid certificate file, cannot append it to the"
                " certificate {}.\nThe error returned by the x509 module was:\n{}".format(
                    append_file, name, str(e)
                )
            )
            return ret

    file_args, extra_args = _get_file_args(name, **kwargs)
    file_args["contents"] = contents

    ret = _certificate_file_managed(ret, file_args)

    if ret["result"]:
        ret["changes"]["Certificate"] = {
            "Old": current_cert_info,
            "New": __salt__["x509.read_certificate"](certificate=name),
        }
        ret["changes"]["Status"] = {
            "Old": invalid_reason,
            "New": "Certificate is valid and up to date",
        }

    return ret


def crl_managed(
    name,
    signing_private_key,
    signing_private_key_passphrase=None,
    signing_cert=None,
    revoked=None,
    days_valid=100,
    digest="",
    days_remaining=30,
    include_expired=False,
    **kwargs
):
    """
    Manage a Certificate Revocation List

    name
        Path to the certificate

    signing_private_key
        The private key that will be used to sign the CRL. This is
        usually your CA's private key.

    signing_private_key_passphrase
        Passphrase to decrypt the private key.

    signing_cert
        The certificate of the authority that will be used to sign the CRL.
        This is usually your CA's certificate.

    revoked
        A list of certificates to revoke. Must include either a serial number or a
        the certificate itself. Can optionally include the revocation date and
        notAfter date from the certificate. See example below for details.

    days_valid : 100
        The number of days the certificate should be valid for.

    digest
        The digest to use for signing the CRL. This has no effect on versions
        of pyOpenSSL less than 0.14.

    days_remaining : 30
        The CRL should be automatically recreated if there are less than
        ``days_remaining`` days until the CRL expires. Set to 0 to disable
        automatic renewal.

    include_expired : False
        If ``True``, include expired certificates in the CRL.

    kwargs
        Any arguments supported by :py:func:`file.managed <salt.states.file.managed>` are supported.

    Example:

    .. code-block:: yaml

        /etc/pki/ca.crl:
          x509.crl_managed:
            - signing_private_key: /etc/pki/myca.key
            - signing_cert: /etc/pki/myca.crt
            - revoked:
              - compromized_Web_key:
                - certificate: /etc/pki/certs/badweb.crt
                - revocation_date: 2015-03-01 00:00:00
                - reason: keyCompromise
              - terminated_vpn_user:
                - serial_number: D6:D2:DC:D8:4D:5C:C0:F4
                - not_after: 2016-01-01 00:00:00
                - revocation_date: 2015-02-25 00:00:00
                - reason: cessationOfOperation
    """
    if revoked is None:
        revoked = []

    revoked = _revoked_to_list(revoked)

    current_days_remaining = 0
    current_comp = {}

    if os.path.isfile(name):
        try:
            current = __salt__["x509.read_crl"](crl=name)
            current_comp = current.copy()
            current_comp.pop("Last Update")
            current_notafter = current_comp.pop("Next Update")
            current_days_remaining = (
                datetime.datetime.strptime(current_notafter, "%Y-%m-%d %H:%M:%S")
                - datetime.datetime.now()
            ).days
            if days_remaining == 0:
                days_remaining = current_days_remaining - 1
        except salt.exceptions.SaltInvocationError:
            current = "{} is not a valid CRL.".format(name)
    else:
        current = "{} does not exist.".format(name)

    new_crl = __salt__["x509.create_crl"](
        text=True,
        signing_private_key=signing_private_key,
        signing_private_key_passphrase=signing_private_key_passphrase,
        signing_cert=signing_cert,
        revoked=revoked,
        days_valid=days_valid,
        digest=digest,
        include_expired=include_expired,
    )

    new = __salt__["x509.read_crl"](crl=new_crl)
    new_comp = new.copy()
    new_comp.pop("Last Update")
    new_comp.pop("Next Update")

    file_args, kwargs = _get_file_args(name, **kwargs)
    new_crl_created = False
    if (
        current_comp == new_comp
        and current_days_remaining > days_remaining
        and __salt__["x509.verify_crl"](name, signing_cert)
    ):
        file_args["contents"] = __salt__["x509.get_pem_entry"](
            name, pem_type="X509 CRL"
        )
    else:
        new_crl_created = True
        file_args["contents"] = new_crl

    ret = __states__["file.managed"](**file_args)
    if new_crl_created:
        ret["changes"] = {"Old": current, "New": __salt__["x509.read_crl"](crl=new_crl)}
    return ret


def pem_managed(name, text, backup=False, **kwargs):
    """
    Manage the contents of a PEM file directly with the content in text, ensuring correct formatting.

    name:
        The path to the file to manage

    text:
        The PEM formatted text to write.

    kwargs:
        Any arguments supported by :py:func:`file.managed <salt.states.file.managed>` are supported.
    """
    file_args, kwargs = _get_file_args(name, **kwargs)
    file_args["contents"] = __salt__["x509.get_pem_entry"](text=text)

    return __states__["file.managed"](**file_args)



"""
This module is used to manage Wordpress installations

:depends: wp binary from http://wp-cli.org/
"""

# Import Python Modules

import collections

# Import Salt Modules
import salt.utils.path

Plugin = collections.namedtuple("Plugin", "name status update versino")


def __virtual__():
    if salt.utils.path.which("wp"):
        return True
    return (False, "Missing dependency: wp")


def _get_plugins(stuff):
    return Plugin(stuff)


def list_plugins(path, user):
    """
    List plugins in an installed wordpress path

    path
        path to wordpress install location

    user
        user to run the command as

    CLI Example:

    .. code-block:: bash

        salt '*' wordpress.list_plugins /var/www/html apache
    """
    ret = []
    resp = __salt__["cmd.shell"]("wp --path={} plugin list".format(path), runas=user)
    for line in resp.split("\n")[1:]:
        ret.append(line.split("\t"))
    return [plugin.__dict__ for plugin in map(_get_plugins, ret)]


def show_plugin(name, path, user):
    """
    Show a plugin in a wordpress install and check if it is installed

    name
        Wordpress plugin name

    path
        path to wordpress install location

    user
        user to run the command as

    CLI Example:

    .. code-block:: bash

        salt '*' wordpress.show_plugin HyperDB /var/www/html apache
    """
    ret = {"name": name}
    resp = __salt__["cmd.shell"](
        "wp --path={} plugin status {}".format(path, name), runas=user
    ).split("\n")
    for line in resp:
        if "Status" in line:
            ret["status"] = line.split(" ")[-1].lower()
        elif "Version" in line:
            ret["version"] = line.split(" ")[-1].lower()
    return ret


def activate(name, path, user):
    """
    Activate a wordpress plugin

    name
        Wordpress plugin name

    path
        path to wordpress install location

    user
        user to run the command as

    CLI Example:

    .. code-block:: bash

        salt '*' wordpress.activate HyperDB /var/www/html apache
    """
    check = show_plugin(name, path, user)
    if check["status"] == "active":
        # already active
        return None
    resp = __salt__["cmd.shell"](
        "wp --path={} plugin activate {}".format(path, name), runas=user
    )
    if "Success" in resp:
        return True
    elif show_plugin(name, path, user)["status"] == "active":
        return True
    return False


def deactivate(name, path, user):
    """
    Deactivate a wordpress plugin

    name
        Wordpress plugin name

    path
        path to wordpress install location

    user
        user to run the command as

    CLI Example:

    .. code-block:: bash

        salt '*' wordpress.deactivate HyperDB /var/www/html apache
    """
    check = show_plugin(name, path, user)
    if check["status"] == "inactive":
        # already inactive
        return None
    resp = __salt__["cmd.shell"](
        "wp --path={} plugin deactivate {}".format(path, name), runas=user
    )
    if "Success" in resp:
        return True
    elif show_plugin(name, path, user)["status"] == "inactive":
        return True
    return False


def is_installed(path, user=None):
    """
    Check if wordpress is installed and setup

    path
        path to wordpress install location

    user
        user to run the command as

    CLI Example:

    .. code-block:: bash

        salt '*' wordpress.is_installed /var/www/html apache
    """
    retcode = __salt__["cmd.retcode"](
        "wp --path={} core is-installed".format(path), runas=user
    )
    if retcode == 0:
        return True
    return False


def install(path, user, admin_user, admin_password, admin_email, title, url):
    """
    Run the initial setup functions for a wordpress install

    path
        path to wordpress install location

    user
        user to run the command as

    admin_user
        Username for the Administrative user for the wordpress install

    admin_password
        Initial Password for the Administrative user for the wordpress install

    admin_email
        Email for the Administrative user for the wordpress install

    title
        Title of the wordpress website for the wordpress install

    url
        Url for the wordpress install

    CLI Example:

    .. code-block:: bash

        salt '*' wordpress.install /var/www/html apache dwallace password123 \
            dwallace@example.com "Daniel's Awesome Blog" https://blog.dwallace.com
    """
    retcode = __salt__["cmd.retcode"](
        'wp --path={} core install --title="{}" --admin_user={} '
        "--admin_password='{}' --admin_email={} --url={}".format(
            path, title, admin_user, admin_password, admin_email, url
        ),
        runas=user,
    )

    if retcode == 0:
        return True
    return False



"""
Module for running imgadm command on SmartOS
"""

import logging

import salt.utils.json
import salt.utils.path
import salt.utils.platform

log = logging.getLogger(__name__)

# Function aliases
__func_alias__ = {
    "list_installed": "list",
    "update_installed": "update",
    "import_image": "import",
}

# Define the module's virtual name
__virtualname__ = "imgadm"


def __virtual__():
    """
    Provides imgadm only on SmartOS
    """
    if salt.utils.platform.is_smartos_globalzone() and salt.utils.path.which("imgadm"):
        return __virtualname__
    return (
        False,
        "{} module can only be loaded on SmartOS compute nodes".format(__virtualname__),
    )


def _exit_status(retcode, stderr=None):
    """
    Translate exit status of imgadm
    """
    ret = {
        0: "Successful completion.",
        1: "An error occurred." if not stderr else stderr,
        2: "Usage error.",
        3: "Image not installed.",
    }[retcode]
    return ret


def _parse_image_meta(image=None, detail=False):
    ret = None

    if image and "Error" in image:
        ret = image
    elif image and "manifest" in image and "name" in image["manifest"]:
        name = image["manifest"]["name"]
        version = image["manifest"]["version"]
        os = image["manifest"]["os"]
        description = image["manifest"]["description"]
        published = image["manifest"]["published_at"]
        source = image["source"]
        if image["manifest"]["name"] == "docker-layer":
            # NOTE: skip docker-layer unless it has a docker:repo and docker:tag
            name = None
            docker_repo = None
            docker_tag = None
            for tag in image["manifest"]["tags"]:
                if tag.startswith("docker:tag:") and image["manifest"]["tags"][tag]:
                    docker_tag = tag.split(":")[-1]
                elif tag == "docker:repo":
                    docker_repo = image["manifest"]["tags"][tag]

            if docker_repo and docker_tag:
                name = "{}:{}".format(docker_repo, docker_tag)
                description = (
                    "Docker image imported from {repo}:{tag} on {date}.".format(
                        repo=docker_repo,
                        tag=docker_tag,
                        date=published,
                    )
                )

        if name and detail:
            ret = {
                "name": name,
                "version": version,
                "os": os,
                "description": description,
                "published": published,
                "source": source,
            }
        elif name:
            ret = "{name}@{version} [{published}]".format(
                name=name,
                version=version,
                published=published,
            )
    else:
        log.debug("smartos_image - encountered invalid image payload: %s", image)
        ret = {"Error": "This looks like an orphaned image, image payload was invalid."}

    return ret


def _split_docker_uuid(uuid):
    """
    Split a smartos docker uuid into repo and tag
    """
    if uuid:
        uuid = uuid.split(":")
        if len(uuid) == 2:
            tag = uuid[1]
            repo = uuid[0]
            return repo, tag
    return None, None


def _is_uuid(uuid):
    """
    Check if uuid is a valid smartos uuid

    Example: e69a0918-055d-11e5-8912-e3ceb6df4cf8
    """
    if uuid and list(len(x) for x in uuid.split("-")) == [8, 4, 4, 4, 12]:
        return True
    return False


def _is_docker_uuid(uuid):
    """
    Check if uuid is a valid smartos docker uuid

    Example plexinc/pms-docker:plexpass
    """
    repo, tag = _split_docker_uuid(uuid)
    return not (not repo and not tag)


def version():
    """
    Return imgadm version

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.version
    """
    ret = {}
    cmd = "imgadm --version"
    res = __salt__["cmd.run"](cmd).splitlines()
    ret = res[0].split()
    return ret[-1]


def docker_to_uuid(uuid):
    """
    Get the image uuid from an imported docker image

    .. versionadded:: 2019.2.0
    """
    if _is_uuid(uuid):
        return uuid
    if _is_docker_uuid(uuid):
        images = list_installed(verbose=True)
        for image_uuid in images:
            if "name" not in images[image_uuid]:
                continue
            if images[image_uuid]["name"] == uuid:
                return image_uuid
    return None


def update_installed(uuid=""):
    """
    Gather info on unknown image(s) (locally installed)

    uuid : string
        optional uuid of image

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.update [uuid]
    """
    cmd = "imgadm update {}".format(uuid).rstrip()
    __salt__["cmd.run"](cmd)
    return {}


def avail(search=None, verbose=False):
    """
    Return a list of available images

    search : string
        search keyword
    verbose : boolean (False)
        toggle verbose output

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.avail [percona]
        salt '*' imgadm.avail verbose=True
    """
    ret = {}
    cmd = "imgadm avail -j"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret

    for image in salt.utils.json.loads(res["stdout"]):
        if image["manifest"]["disabled"] or not image["manifest"]["public"]:
            continue
        if search and search not in image["manifest"]["name"]:
            # we skip if we are searching but don't have a match
            continue
        uuid = image["manifest"]["uuid"]
        data = _parse_image_meta(image, verbose)
        if data:
            ret[uuid] = data

    return ret


def list_installed(verbose=False):
    """
    Return a list of installed images

    verbose : boolean (False)
        toggle verbose output

    .. versionchanged:: 2019.2.0

        Docker images are now also listed

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.list
        salt '*' imgadm.list docker=True
        salt '*' imgadm.list verbose=True
    """
    ret = {}
    cmd = "imgadm list -j"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret

    for image in salt.utils.json.loads(res["stdout"]):
        uuid = image["manifest"]["uuid"]
        data = _parse_image_meta(image, verbose)
        if data:
            ret[uuid] = data

    return ret


def show(uuid):
    """
    Show manifest of a given image

    uuid : string
        uuid of image

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.show e42f8c84-bbea-11e2-b920-078fab2aab1f
        salt '*' imgadm.show plexinc/pms-docker:plexpass
    """
    ret = {}

    if _is_uuid(uuid) or _is_docker_uuid(uuid):
        cmd = "imgadm show {}".format(uuid)
        res = __salt__["cmd.run_all"](cmd, python_shell=False)
        retcode = res["retcode"]
        if retcode != 0:
            ret["Error"] = _exit_status(retcode, res["stderr"])
        else:
            ret = salt.utils.json.loads(res["stdout"])
    else:
        ret["Error"] = "{} is not a valid uuid.".format(uuid)

    return ret


def get(uuid):
    """
    Return info on an installed image

    uuid : string
        uuid of image

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.get e42f8c84-bbea-11e2-b920-078fab2aab1f
        salt '*' imgadm.get plexinc/pms-docker:plexpass
    """
    ret = {}

    if _is_docker_uuid(uuid):
        uuid = docker_to_uuid(uuid)

    if _is_uuid(uuid):
        cmd = "imgadm get {}".format(uuid)
        res = __salt__["cmd.run_all"](cmd, python_shell=False)
        retcode = res["retcode"]
        if retcode != 0:
            ret["Error"] = _exit_status(retcode, res["stderr"])
        else:
            ret = salt.utils.json.loads(res["stdout"])
    else:
        ret["Error"] = "{} is not a valid uuid.".format(uuid)

    return ret


def import_image(uuid, verbose=False):
    """
    Import an image from the repository

    uuid : string
        uuid to import
    verbose : boolean (False)
        toggle verbose output

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.import e42f8c84-bbea-11e2-b920-078fab2aab1f [verbose=True]
    """
    ret = {}
    cmd = "imgadm import {}".format(uuid)
    res = __salt__["cmd.run_all"](cmd, python_shell=False)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret

    uuid = docker_to_uuid(uuid)
    data = _parse_image_meta(get(uuid), verbose)
    return {uuid: data}


def delete(uuid):
    """
    Remove an installed image

    uuid : string
        Specifies uuid to import

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.delete e42f8c84-bbea-11e2-b920-078fab2aab1f
    """
    ret = {}
    cmd = "imgadm delete {}".format(uuid)
    res = __salt__["cmd.run_all"](cmd, python_shell=False)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret
    # output: Deleted image d5b3865c-0804-11e5-be21-dbc4ce844ddc
    result = []
    for image in res["stdout"].splitlines():
        image = [var for var in image.split(" ") if var]
        result.append(image[2])

    return result


def vacuum(verbose=False):
    """
    Remove unused images

    verbose : boolean (False)
        toggle verbose output

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.vacuum [verbose=True]
    """
    ret = {}
    cmd = "imgadm vacuum -f"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret
    # output: Deleted image d5b3865c-0804-11e5-be21-dbc4ce844ddc (lx-centos-6@20150601)
    result = {}
    for image in res["stdout"].splitlines():
        image = [var for var in image.split(" ") if var]
        result[image[2]] = {
            "name": image[3][1 : image[3].index("@")],
            "version": image[3][image[3].index("@") + 1 : -1],
        }
    if verbose:
        return result
    else:
        return list(result.keys())


def sources(verbose=False):
    """
    Return a list of available sources

    verbose : boolean (False)
        toggle verbose output

    .. versionadded:: 2019.2.0

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.sources
    """
    ret = {}
    cmd = "imgadm sources -j"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret

    for src in salt.utils.json.loads(res["stdout"]):
        ret[src["url"]] = src
        del src["url"]

    if not verbose:
        ret = list(ret)

    return ret


def source_delete(source):
    """
    Delete a source

    source : string
        source url to delete

    .. versionadded:: 2019.2.0

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.source_delete https://updates.joyent.com
    """
    ret = {}
    cmd = "imgadm sources -d {}".format(source)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode, res["stderr"])
        return ret

    return sources(False)


def source_add(source, source_type="imgapi"):
    """
    Add a new source

    source : string
        source url to add
    source_trype : string (imgapi)
        source type, either imgapi or docker

    .. versionadded:: 2019.2.0

    CLI Example:

    .. code-block:: bash

        salt '*' imgadm.source_add https://updates.joyent.com
        salt '*' imgadm.source_add https://docker.io docker
    """
    ret = {}

    # NOTE: there are some undocumented deprecated source types
    #       so we just warn instead of error on those
    if source_type not in ["imgapi", "docker"]:
        log.warning("Possible unsupported imgage source type specified!")

    cmd = "imgadm sources -a {} -t {}".format(source, source_type)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode, res["stderr"])
        return ret

    return sources(False)


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



import base64

import pytest
import salt.modules.cmdmod as cmdmod
import salt.utils.path
import salt.utils.stringutils

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(params=["powershell", "pwsh"])
def shell(request):
    """
    This will run the test on powershell and powershell core (pwsh). If
    powershell core is not installed that test run will be skipped
    """
    if request.param == "pwsh" and salt.utils.path.which("pwsh") is None:
        pytest.skip("Powershell 7 Not Present")
    return request.param


def test_powershell(shell):
    """
    Test cmd.powershell
    """
    ret = cmdmod.powershell("Write-Output foo", shell=shell)
    assert ret == "foo"


def test_powershell_encode_cmd(shell):
    """
    Test cmd.powershell with encode_cmd
    """
    ret = cmdmod.powershell('Write-Output "encoded foo"', encode_cmd=True, shell=shell)
    assert ret == "encoded foo"


def test_powershell_all(shell):
    """
    Test cmd.powershell_all
    """
    ret = cmdmod.powershell_all("Write-Output foo", shell=shell)
    assert isinstance(ret["pid"], int)
    assert ret["retcode"] == 0
    assert ret["stderr"] == ""
    assert ret["result"] == "foo"


def test_powershell_all_encode_cmd(shell):
    """
    Test cmd.powershell_all with encode_cmd
    """
    ret = cmdmod.powershell_all(
        'Write-Output "encoded foo"', encode_cmd=True, shell=shell
    )
    assert isinstance(ret["pid"], int)
    assert ret["retcode"] == 0
    assert ret["stderr"] == ""
    assert ret["result"] == "encoded foo"


def test_cmd_run_all_powershell_list():
    """
    Ensure that cmd.run_all supports running shell='powershell' with cmd passed
    as a list
    """
    ret = cmdmod.run_all(
        cmd=["Write-Output", "salt"], python_shell=False, shell="powershell"
    )
    assert ret["stdout"] == "salt"


def test_cmd_run_all_powershell_string():
    """
    Ensure that cmd.run_all supports running shell='powershell' with cmd passed
     as a string
    """
    ret = cmdmod.run_all(
        cmd="Write-Output salt", python_shell=False, shell="powershell"
    )
    assert ret["stdout"] == "salt"


def test_cmd_run_encoded_cmd(shell):
    cmd = "Write-Output 'encoded command'"
    cmd = "$ProgressPreference='SilentlyContinue'; {}".format(cmd)
    cmd_utf16 = cmd.encode("utf-16-le")
    encoded_cmd = base64.standard_b64encode(cmd_utf16)
    encoded_cmd = salt.utils.stringutils.to_str(encoded_cmd)
    ret = cmdmod.run(cmd=encoded_cmd, shell=shell, encoded_cmd=True)
    assert ret == "encoded command"


def test_cmd_run_all_encoded_cmd(shell):
    cmd = "Write-Output 'encoded command'"
    cmd = "$ProgressPreference='SilentlyContinue'; {}".format(cmd)
    cmd_utf16 = cmd.encode("utf-16-le")
    encoded_cmd = base64.standard_b64encode(cmd_utf16)
    encoded_cmd = salt.utils.stringutils.to_str(encoded_cmd)
    ret = cmdmod.run_all(cmd=encoded_cmd, shell=shell, encoded_cmd=True)
    assert ret["stdout"] == "encoded command"



import logging

import salt.utils.dicttrim as dicttrimmer
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class DictTrimTestCase(TestCase):
    def setUp(self):
        self.old_dict = {"a": "b", "c": "x" * 10000}
        self.new_dict = {"a": "b", "c": "VALUE_TRIMMED"}

    def test_trim_dict(self):
        ret = dicttrimmer.trim_dict(self.old_dict, 1000)
        self.assertEqual(ret, self.new_dict)


class RecursiveDictTrimTestCase(TestCase):
    def setUp(self):
        self.old_dict = {"a": {"b": 1, "c": 2, "e": "x" * 10000, "f": "3"}}
        self.new_dict = {"a": {"b": 1, "c": 2, "e": "VALUE_TRIMMED", "f": "3"}}

    def test_trim_dict(self):
        ret = dicttrimmer.trim_dict(self.old_dict, 1000)
        self.assertEqual(ret, self.new_dict)



"""
    :codeauthor: Eric Radman <ericshane@eradman.com>
"""


from salt.cloud.clouds import joyent
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


# Stubs
def fake_wait_for_ip(
    check_for_ip_fn, interval=None, timeout=None, interval_multiplier=None
):
    """
    Callback that returns immediately instead of waiting
    """
    assert isinstance(interval, int)
    assert isinstance(timeout, int)
    assert isinstance(interval_multiplier, int)
    return check_for_ip_fn()


@skipIf(
    joyent.HAS_REQUIRED_CRYPTO is False, reason="PyCrypto or Cryptodome not installed"
)
class JoyentTestCase(TestCase, LoaderModuleMockMixin):
    """
    Unit TestCase for the salt.cloud.clouds.joyent module
    """

    def setup_loader_modules(self):
        patcher = patch("salt.utils.cloud.wait_for_ip", fake_wait_for_ip)
        patcher.start()
        self.addCleanup(patcher.stop)
        return {
            joyent: {
                "__utils__": {
                    "cloud.fire_event": MagicMock(),
                    "cloud.bootstrap": MagicMock(),
                },
                "__opts__": {
                    "sock_dir": True,
                    "transport": True,
                    "providers": {"my_joyent": {}},
                    "profiles": {"my_joyent": {}},
                },
                "__active_provider_name__": "my_joyent:joyent",
            }
        }

    def setUp(self):
        self.vm_ = {
            "profile": "my_joyent",
            "name": "vm3",
            "driver": "joyent",
            "size": "k4-highcpu-kvm-750M",
            "image": "freebsd10",
            "location": "us-east-1",
        }

    def tearDown(self):
        del self.vm_

    def test_query_instance_init(self):
        """
        Initial provisioning, no IP assigned
        """
        # Not yet reachable
        reply = (200, {"state": "provisioning"})
        with patch.object(joyent, "show_instance", return_value=reply):
            result = joyent.query_instance(self.vm_)
        self.assertTrue(joyent.__utils__["cloud.fire_event"].called_once())
        self.assertEqual(result, None)

    def test_query_instance_has_ip(self):
        """
        IP address assigned but not yet ready
        """
        reply = (200, {"primaryIp": "1.1.1.1", "state": "provisioning"})
        with patch.object(joyent, "show_instance", return_value=reply):
            result = joyent.query_instance(self.vm_)
        self.assertTrue(joyent.__utils__["cloud.fire_event"].called_once())
        self.assertEqual(result, None)

    def test_query_instance_ready(self):
        """
        IP address assigned, and VM is ready
        """
        reply = (200, {"primaryIp": "1.1.1.1", "state": "running"})
        with patch.object(joyent, "show_instance", return_value=reply):
            result = joyent.query_instance(self.vm_)
        self.assertTrue(joyent.__utils__["cloud.fire_event"].called_once())
        self.assertEqual(result, "1.1.1.1")



"""
    tests.integration.shell.auth
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import logging

import pytest
import salt.utils.platform
import salt.utils.pycrypto

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.destructive_test,
    pytest.mark.skip_on_windows,
]

USERA = "saltdev-auth"
USERA_PWD = "saltdev"


@pytest.fixture(scope="module")
def saltdev_account(sminion):
    try:
        assert sminion.functions.user.add(USERA, createhome=False)
        assert sminion.functions.shadow.set_password(
            USERA,
            USERA_PWD
            if salt.utils.platform.is_darwin()
            else salt.utils.pycrypto.gen_hash(password=USERA_PWD),
        )
        assert USERA in sminion.functions.user.list_users()
        # Run tests
        yield
    finally:
        sminion.functions.user.delete(USERA, remove=True)


SALTOPS = "saltops"


@pytest.fixture(scope="module")
def saltops_group(sminion):
    try:
        assert sminion.functions.group.add(SALTOPS)
        # Run tests
        yield
    finally:
        sminion.functions.group.delete(SALTOPS)


USERB = "saltdev-adm"
USERB_PWD = USERA_PWD


@pytest.fixture(scope="module")
def saltadm_account(sminion, saltops_group):
    try:
        assert sminion.functions.user.add(USERB, groups=[SALTOPS], createhome=False)
        assert sminion.functions.shadow.set_password(
            USERB,
            USERB_PWD
            if salt.utils.platform.is_darwin()
            else salt.utils.pycrypto.gen_hash(password=USERB_PWD),
        )
        assert USERB in sminion.functions.user.list_users()
        # Run tests
        yield
    finally:
        sminion.functions.user.delete(USERB, remove=True)


def test_pam_auth_valid_user(salt_minion, salt_cli, saltdev_account):
    """
    test that pam auth mechanism works with a valid user
    """
    # test user auth against pam
    ret = salt_cli.run(
        "-a",
        "pam",
        "--username",
        USERA,
        "--password",
        USERA_PWD,
        "test.ping",
        minion_tgt=salt_minion.id,
    )
    assert ret.exitcode == 0
    assert ret.json is True


def test_pam_auth_invalid_user(salt_minion, salt_cli, saltdev_account):
    """
    test pam auth mechanism errors for an invalid user
    """
    ret = salt_cli.run(
        "-a",
        "pam",
        "--username",
        "nouser",
        "--password",
        "1234",
        "test.ping",
        minion_tgt=salt_minion.id,
    )
    assert ret.stdout == "Authentication error occurred."


def test_pam_auth_valid_group(salt_minion, salt_cli, saltadm_account):
    """
    test that pam auth mechanism works for a valid group
    """
    # test group auth against pam: saltadm is not configured in
    # external_auth, but saltops is and saldadm is a member of saltops
    ret = salt_cli.run(
        "-a",
        "pam",
        "--username",
        USERB,
        "--password",
        USERB_PWD,
        "test.ping",
        minion_tgt=salt_minion.id,
    )
    assert ret.exitcode == 0
    assert ret.json is True



"""
StatusPage
==========

Manage the StatusPage_ configuration.

.. _StatusPage: https://www.statuspage.io/

In the minion configuration file, the following block is required:

.. code-block:: yaml

  statuspage:
    api_key: <API_KEY>
    page_id: <PAGE_ID>

.. versionadded:: 2017.7.0
"""

import logging
import time

# ----------------------------------------------------------------------------------------------------------------------
# module properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "statuspage"

log = logging.getLogger(__file__)

_DO_NOT_COMPARE_FIELDS = ["created_at", "updated_at"]

_MATCH_KEYS = ["id", "name"]

_PACE = 1  # 1 request per second

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    Return the execution module virtualname.
    """
    if "statuspage.create" in __salt__:
        return True
    return (False, "statuspage module could not be loaded")


def _default_ret(name):
    """
    Default dictionary returned.
    """
    return {"name": name, "result": False, "comment": "", "changes": {}}


def _compute_diff_ret():
    """
    Default dictionary retuned by the _compute_diff helper.
    """
    return {"add": [], "update": [], "remove": []}


def _clear_dict(endpoint_props):
    """
    Eliminates None entries from the features of the endpoint dict.
    """
    return {
        prop_name: prop_val
        for prop_name, prop_val in endpoint_props.items()
        if prop_val is not None
    }


def _ignore_keys(endpoint_props):
    """
    Ignores some keys that might be different without any important info.
    These keys are defined under _DO_NOT_COMPARE_FIELDS.
    """
    return {
        prop_name: prop_val
        for prop_name, prop_val in endpoint_props.items()
        if prop_name not in _DO_NOT_COMPARE_FIELDS
    }


def _unique(list_of_dicts):
    """
    Returns an unique list of dictionaries given a list that may contain duplicates.
    """
    unique_list = []
    for ele in list_of_dicts:
        if ele not in unique_list:
            unique_list.append(ele)
    return unique_list


def _clear_ignore(endpoint_props):
    """
    Both _clear_dict and _ignore_keys in a single iteration.
    """
    return {
        prop_name: prop_val
        for prop_name, prop_val in endpoint_props.items()
        if prop_name not in _DO_NOT_COMPARE_FIELDS and prop_val is not None
    }


def _clear_ignore_list(lst):
    """
    Apply _clear_ignore to a list.
    """
    return _unique([_clear_ignore(ele) for ele in lst])


def _find_match(ele, lst):
    """
    Find a matching element in a list.
    """
    for _ele in lst:
        for match_key in _MATCH_KEYS:
            if _ele.get(match_key) == ele.get(match_key):
                return ele


def _update_on_fields(prev_ele, new_ele):
    """
    Return a dict with fields that differ between two dicts.
    """
    fields_update = {
        prop_name: prop_val
        for prop_name, prop_val in new_ele.items()
        if new_ele.get(prop_name) != prev_ele.get(prop_name) or prop_name in _MATCH_KEYS
    }
    if len(set(fields_update.keys()) | set(_MATCH_KEYS)) > len(set(_MATCH_KEYS)):
        if "id" not in fields_update:
            # in case of update, the ID is necessary
            # if not specified in the pillar,
            # will try to get it from the prev_ele
            fields_update["id"] = prev_ele["id"]
        return fields_update


def _compute_diff(expected_endpoints, configured_endpoints):
    """
    Compares configured endpoints with the expected configuration and returns the differences.
    """
    new_endpoints = []
    update_endpoints = []
    remove_endpoints = []

    ret = _compute_diff_ret()

    # noth configured => configure with expected endpoints
    if not configured_endpoints:
        ret.update({"add": expected_endpoints})
        return ret

    # noting expected => remove everything
    if not expected_endpoints:
        ret.update({"remove": configured_endpoints})
        return ret

    expected_endpoints_clear = _clear_ignore_list(expected_endpoints)
    configured_endpoints_clear = _clear_ignore_list(configured_endpoints)

    for expected_endpoint_clear in expected_endpoints_clear:
        if expected_endpoint_clear not in configured_endpoints_clear:
            # none equal => add or update
            matching_ele = _find_match(
                expected_endpoint_clear, configured_endpoints_clear
            )
            if not matching_ele:
                # new element => add
                new_endpoints.append(expected_endpoint_clear)
            else:
                # element matched, but some fields are different
                update_fields = _update_on_fields(matching_ele, expected_endpoint_clear)
                if update_fields:
                    update_endpoints.append(update_fields)
    for configured_endpoint_clear in configured_endpoints_clear:
        if configured_endpoint_clear not in expected_endpoints_clear:
            matching_ele = _find_match(
                configured_endpoint_clear, expected_endpoints_clear
            )
            if not matching_ele:
                #  no match found => remove
                remove_endpoints.append(configured_endpoint_clear)

    return {
        "add": new_endpoints,
        "update": update_endpoints,
        "remove": remove_endpoints,
    }


# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


def create(
    name,
    endpoint="incidents",
    api_url=None,
    page_id=None,
    api_key=None,
    api_version=None,
    **kwargs
):
    """
    Insert a new entry under a specific endpoint.

    endpoint: incidents
        Insert under this specific endpoint.

    page_id
        Page ID. Can also be specified in the config file.

    api_key
        API key. Can also be specified in the config file.

    api_version: 1
        API version. Can also be specified in the config file.

    api_url
        Custom API URL in case the user has a StatusPage service running in a custom environment.

    kwargs
        Other params.

    SLS Example:

    .. code-block:: yaml

        create-my-component:
            statuspage.create:
                - endpoint: components
                - name: my component
                - group_id: 993vgplshj12
    """
    ret = _default_ret(name)
    endpoint_sg = endpoint[:-1]  # singular
    if __opts__["test"]:
        ret["comment"] = "The following {endpoint} would be created:".format(
            endpoint=endpoint_sg
        )
        ret["result"] = None
        ret["changes"][endpoint] = {}
        for karg, warg in kwargs.items():
            if warg is None or karg.startswith("__"):
                continue
            ret["changes"][endpoint][karg] = warg
        return ret
    sp_create = __salt__["statuspage.create"](
        endpoint=endpoint,
        api_url=api_url,
        page_id=page_id,
        api_key=api_key,
        api_version=api_version,
        **kwargs
    )
    if not sp_create.get("result"):
        ret["comment"] = "Unable to create {endpoint}: {msg}".format(
            endpoint=endpoint_sg, msg=sp_create.get("comment")
        )
    else:
        ret["comment"] = "{endpoint} created!".format(endpoint=endpoint_sg)
        ret["result"] = True
        ret["changes"] = sp_create.get("out")


def update(
    name,
    endpoint="incidents",
    id=None,
    api_url=None,
    page_id=None,
    api_key=None,
    api_version=None,
    **kwargs
):
    """
    Update attribute(s) of a specific endpoint.

    id
        The unique ID of the endpoint entry.

    endpoint: incidents
        Endpoint name.

    page_id
        Page ID. Can also be specified in the config file.

    api_key
        API key. Can also be specified in the config file.

    api_version: 1
        API version. Can also be specified in the config file.

    api_url
        Custom API URL in case the user has a StatusPage service running in a custom environment.

    SLS Example:

    .. code-block:: yaml

        update-my-incident:
            statuspage.update:
                - id: dz959yz2nd4l
                - status: resolved
    """
    ret = _default_ret(name)
    endpoint_sg = endpoint[:-1]  # singular
    if not id:
        log.error("Invalid %s ID", endpoint_sg)
        ret["comment"] = "Please specify a valid {endpoint} ID".format(
            endpoint=endpoint_sg
        )
        return ret
    if __opts__["test"]:
        ret["comment"] = "{endpoint} #{id} would be updated:".format(
            endpoint=endpoint_sg, id=id
        )
        ret["result"] = None
        ret["changes"][endpoint] = {}
        for karg, warg in kwargs.items():
            if warg is None or karg.startswith("__"):
                continue
            ret["changes"][endpoint][karg] = warg
        return ret
    sp_update = __salt__["statuspage.update"](
        endpoint=endpoint,
        id=id,
        api_url=api_url,
        page_id=page_id,
        api_key=api_key,
        api_version=api_version,
        **kwargs
    )
    if not sp_update.get("result"):
        ret["comment"] = "Unable to update {endpoint} #{id}: {msg}".format(
            endpoint=endpoint_sg, id=id, msg=sp_update.get("comment")
        )
    else:
        ret["comment"] = "{endpoint} #{id} updated!".format(endpoint=endpoint_sg, id=id)
        ret["result"] = True
        ret["changes"] = sp_update.get("out")


def delete(
    name,
    endpoint="incidents",
    id=None,
    api_url=None,
    page_id=None,
    api_key=None,
    api_version=None,
):
    """
    Remove an entry from an endpoint.

    endpoint: incidents
        Request a specific endpoint.

    page_id
        Page ID. Can also be specified in the config file.

    api_key
        API key. Can also be specified in the config file.

    api_version: 1
        API version. Can also be specified in the config file.

    api_url
        Custom API URL in case the user has a StatusPage service running in a custom environment.

    SLS Example:

    .. code-block:: yaml

        delete-my-component:
            statuspage.delete:
                - endpoint: components
                - id: ftgks51sfs2d
    """
    ret = _default_ret(name)
    endpoint_sg = endpoint[:-1]  # singular
    if not id:
        log.error("Invalid %s ID", endpoint_sg)
        ret["comment"] = "Please specify a valid {endpoint} ID".format(
            endpoint=endpoint_sg
        )
        return ret
    if __opts__["test"]:
        ret["comment"] = "{endpoint} #{id} would be removed!".format(
            endpoint=endpoint_sg, id=id
        )
        ret["result"] = None
    sp_delete = __salt__["statuspage.delete"](
        endpoint=endpoint,
        id=id,
        api_url=api_url,
        page_id=page_id,
        api_key=api_key,
        api_version=api_version,
    )
    if not sp_delete.get("result"):
        ret["comment"] = "Unable to delete {endpoint} #{id}: {msg}".format(
            endpoint=endpoint_sg, id=id, msg=sp_delete.get("comment")
        )
    else:
        ret["comment"] = "{endpoint} #{id} deleted!".format(endpoint=endpoint_sg, id=id)
        ret["result"] = True


def managed(
    name,
    config,
    api_url=None,
    page_id=None,
    api_key=None,
    api_version=None,
    pace=_PACE,
    allow_empty=False,
):
    """
    Manage the StatusPage configuration.

    config
        Dictionary with the expected configuration of the StatusPage.
        The main level keys of this dictionary represent the endpoint name.
        If a certain endpoint does not exist in this structure, it will be ignored / not configured.

    page_id
        Page ID. Can also be specified in the config file.

    api_key
        API key. Can also be specified in the config file.

    api_version: 1
        API version. Can also be specified in the config file.

    api_url
        Custom API URL in case the user has a StatusPage service running in a custom environment.

    pace: 1
        Max requests per second allowed by the API.

    allow_empty: False
        Allow empty config.

    SLS example:

    .. code-block:: yaml

        my-statuspage-config:
            statuspage.managed:
                - config:
                    components:
                        - name: component1
                          group_id: uy4g37rf
                        - name: component2
                          group_id: 3n4uyu4gf
                    incidents:
                        - name: incident1
                          status: resolved
                          impact: major
                          backfilled: false
                        - name: incident2
                          status: investigating
                          impact: minor
    """
    complete_diff = {}
    ret = _default_ret(name)
    if not config and not allow_empty:
        ret.update(
            {
                "result": False,
                "comment": (
                    "Cannot remove everything. To allow this, please set the option"
                    " `allow_empty` as True."
                ),
            }
        )
        return ret
    is_empty = True
    for endpoint_name, endpoint_expected_config in config.items():
        if endpoint_expected_config:
            is_empty = False
        endpoint_existing_config_ret = __salt__["statuspage.retrieve"](
            endpoint=endpoint_name,
            api_url=api_url,
            page_id=page_id,
            api_key=api_key,
            api_version=api_version,
        )
        if not endpoint_existing_config_ret.get("result"):
            ret.update({"comment": endpoint_existing_config_ret.get("comment")})
            return ret  # stop at first error
        endpoint_existing_config = endpoint_existing_config_ret.get("out")
        complete_diff[endpoint_name] = _compute_diff(
            endpoint_expected_config, endpoint_existing_config
        )
    if is_empty and not allow_empty:
        ret.update(
            {
                "result": False,
                "comment": (
                    "Cannot remove everything. To allow this, please set the option"
                    " `allow_empty` as True."
                ),
            }
        )
        return ret
    any_changes = False
    for endpoint_name, endpoint_diff in complete_diff.items():
        if (
            endpoint_diff.get("add")
            or endpoint_diff.get("update")
            or endpoint_diff.get("remove")
        ):
            any_changes = True
    if not any_changes:
        ret.update({"result": True, "comment": "No changes required.", "changes": {}})
        return ret
    ret.update({"changes": complete_diff})
    if __opts__.get("test"):
        ret.update(
            {
                "comment": "Testing mode. Would apply the following changes:",
                "result": None,
            }
        )
        return ret
    for endpoint_name, endpoint_diff in complete_diff.items():
        endpoint_sg = endpoint_name[:-1]  # singular
        for new_endpoint in endpoint_diff.get("add"):
            log.debug("Defining new %s %s", endpoint_sg, new_endpoint)
            adding = __salt__["statuspage.create"](
                endpoint=endpoint_name,
                api_url=api_url,
                page_id=page_id,
                api_key=api_key,
                api_version=api_version,
                **new_endpoint
            )
            if not adding.get("result"):
                ret.update({"comment": adding.get("comment")})
                return ret
            if pace:
                time.sleep(1 / pace)
        for update_endpoint in endpoint_diff.get("update"):
            if "id" not in update_endpoint:
                continue
            endpoint_id = update_endpoint.pop("id")
            log.debug("Updating %s #%s: %s", endpoint_sg, endpoint_id, update_endpoint)
            updating = __salt__["statuspage.update"](
                endpoint=endpoint_name,
                id=endpoint_id,
                api_url=api_url,
                page_id=page_id,
                api_key=api_key,
                api_version=api_version,
                **update_endpoint
            )
            if not updating.get("result"):
                ret.update({"comment": updating.get("comment")})
                return ret
            if pace:
                time.sleep(1 / pace)
        for remove_endpoint in endpoint_diff.get("remove"):
            if "id" not in remove_endpoint:
                continue
            endpoint_id = remove_endpoint.pop("id")
            log.debug("Removing %s #%s", endpoint_sg, endpoint_id)
            removing = __salt__["statuspage.delete"](
                endpoint=endpoint_name,
                id=endpoint_id,
                api_url=api_url,
                page_id=page_id,
                api_key=api_key,
                api_version=api_version,
            )
            if not removing.get("result"):
                ret.update({"comment": removing.get("comment")})
                return ret
            if pace:
                time.sleep(1 / pace)
    ret.update({"result": True, "comment": "StatusPage updated."})
    return ret



"""
A module to pull data from Cobbler via its API into the Pillar dictionary


Configuring the Cobbler ext_pillar
==================================

The same cobbler.* parameters are used for both the Cobbler tops and Cobbler pillar
modules.

.. code-block:: yaml

  ext_pillar:
  - cobbler:
      key: cobbler # Nest results within this key. By default, values are not nested.
      only: [parameters] # Add only these keys to pillar.

  cobbler.url: https://example.com/cobbler_api #default is http://localhost/cobbler_api
  cobbler.user: username # default is no username
  cobbler.password: password # default is no password


Module Documentation
====================
"""

import logging
import xmlrpc.client

__opts__ = {
    "cobbler.url": "http://localhost/cobbler_api",
    "cobbler.user": None,
    "cobbler.password": None,
}


# Set up logging
log = logging.getLogger(__name__)


def ext_pillar(minion_id, pillar, key=None, only=()):  # pylint: disable=W0613
    """
    Read pillar data from Cobbler via its API.
    """
    url = __opts__["cobbler.url"]
    user = __opts__["cobbler.user"]
    password = __opts__["cobbler.password"]

    log.info("Querying cobbler at %r for information for %r", url, minion_id)
    try:
        server = xmlrpc.client.Server(url, allow_none=True)
        if user:
            server.login(user, password)
        result = server.get_blended_data(None, minion_id)
    except Exception:  # pylint: disable=broad-except
        log.exception("Could not connect to cobbler.")
        return {}

    if only:
        result = {k: result[k] for k in only if k in result}

    if key:
        result = {key: result}
    return result



"""
    :synopsis: Unit Tests for Windows iis Module 'state.win_iis'
    :platform: Windows
    .. versionadded:: 2019.2.2
"""

import pytest
import salt.states.win_iis as win_iis
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_iis: {}}


def __base_webconfiguration_ret(comment="", changes=None, name="", result=None):
    return {
        "name": name,
        "changes": changes if changes else {},
        "comment": comment,
        "result": result,
    }


def test_webconfiguration_settings_no_settings():
    name = "IIS"
    settings = {}
    expected_ret = __base_webconfiguration_ret(
        name=name, comment="No settings to change provided.", result=True
    )
    actual_ret = win_iis.webconfiguration_settings(name, settings)
    assert expected_ret == actual_ret


def test_webconfiguration_settings_collection_failure():
    name = "IIS:\\"
    settings = {
        "system.applicationHost/sites": {
            "Collection[{name: site0}].logFile.directory": "C:\\logs\\iis\\site0",
        },
    }
    old_settings = [
        {
            "filter": "system.applicationHost/sites",
            "name": "Collection[{name: site0}].logFile.directory",
            "value": "C:\\logs\\iis\\old_site",
        }
    ]
    current_settings = old_settings
    new_settings = old_settings
    expected_ret = __base_webconfiguration_ret(
        name=name,
        result=False,
        changes={
            "changes": {
                old_settings[0]["filter"]
                + "."
                + old_settings[0]["name"]: {
                    "old": old_settings[0]["value"],
                    "new": settings[old_settings[0]["filter"]][old_settings[0]["name"]],
                }
            },
            "failures": {
                old_settings[0]["filter"]
                + "."
                + old_settings[0]["name"]: {
                    "old": old_settings[0]["value"],
                    "new": new_settings[0]["value"],
                }
            },
        },
        comment="Some settings failed to change.",
    )
    with patch.dict(
        win_iis.__salt__,
        {
            "win_iis.get_webconfiguration_settings": MagicMock(
                side_effect=[old_settings, current_settings, new_settings]
            ),
            "win_iis.set_webconfiguration_settings": MagicMock(return_value=True),
        },
    ), patch.dict(win_iis.__opts__, {"test": False}):
        actual_ret = win_iis.webconfiguration_settings(name, settings)
    assert expected_ret == actual_ret


def test_webconfiguration_settings_collection():
    name = "IIS:\\"
    settings = {
        "system.applicationHost/sites": {
            "Collection[{name: site0}].logFile.directory": "C:\\logs\\iis\\site0",
        },
    }
    old_settings = [
        {
            "filter": "system.applicationHost/sites",
            "name": "Collection[{name: site0}].logFile.directory",
            "value": "C:\\logs\\iis\\old_site",
        }
    ]
    current_settings = [
        {
            "filter": "system.applicationHost/sites",
            "name": "Collection[{name: site0}].logFile.directory",
            "value": "C:\\logs\\iis\\site0",
        }
    ]
    new_settings = current_settings
    expected_ret = __base_webconfiguration_ret(
        name=name,
        result=True,
        changes={
            old_settings[0]["filter"]
            + "."
            + old_settings[0]["name"]: {
                "old": old_settings[0]["value"],
                "new": new_settings[0]["value"],
            }
        },
        comment="Set settings to contain the provided values.",
    )
    with patch.dict(
        win_iis.__salt__,
        {
            "win_iis.get_webconfiguration_settings": MagicMock(
                side_effect=[old_settings, current_settings, new_settings]
            ),
            "win_iis.set_webconfiguration_settings": MagicMock(return_value=True),
        },
    ), patch.dict(win_iis.__opts__, {"test": False}):
        actual_ret = win_iis.webconfiguration_settings(name, settings)
    assert expected_ret == actual_ret


def test_container_settings_password_redacted():
    name = "IIS:\\"
    container = "AppPools"
    settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "Sup3rS3cr3tP@ssW0rd",
        "processModel.identityType": "SpecificUser",
    }
    old_settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "0ldP@ssW0rd1!",
        "processModel.identityType": "SpecificUser",
    }
    current_settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "Sup3rS3cr3tP@ssW0rd",
        "processModel.identityType": "SpecificUser",
    }

    new_settings = current_settings
    expected_ret = {
        "name": name,
        "changes": {
            "processModel.password": {
                "new": "XXX-REDACTED-XXX",
                "old": "XXX-REDACTED-XXX",
            }
        },
        "comment": "Set settings to contain the provided values.",
        "result": True,
    }
    with patch.dict(
        win_iis.__salt__,
        {
            "win_iis.get_container_setting": MagicMock(
                side_effect=[old_settings, current_settings, new_settings]
            ),
            "win_iis.set_container_setting": MagicMock(return_value=True),
        },
    ), patch.dict(win_iis.__opts__, {"test": False}):
        actual_ret = win_iis.container_setting(
            name=name, container=container, settings=settings
        )
    assert expected_ret == actual_ret


def test_container_settings_password_redacted_test_true():
    name = "IIS:\\"
    container = "AppPools"
    settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "Sup3rS3cr3tP@ssW0rd",
        "processModel.identityType": "SpecificUser",
    }
    old_settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "0ldP@ssW0rd1!",
        "processModel.identityType": "SpecificUser",
    }
    current_settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "Sup3rS3cr3tP@ssW0rd",
        "processModel.identityType": "SpecificUser",
    }
    new_settings = current_settings
    expected_ret = {
        "name": name,
        "changes": {
            "processModel.password": {
                "new": "XXX-REDACTED-XXX",
                "old": "XXX-REDACTED-XXX",
            }
        },
        "comment": "Settings will be changed.",
        "result": None,
    }
    with patch.dict(
        win_iis.__salt__,
        {
            "win_iis.get_container_setting": MagicMock(
                side_effect=[old_settings, current_settings, new_settings]
            ),
            "win_iis.set_container_setting": MagicMock(return_value=True),
        },
    ), patch.dict(win_iis.__opts__, {"test": True}):
        actual_ret = win_iis.container_setting(
            name=name, container=container, settings=settings
        )
    assert expected_ret == actual_ret


def test_container_settings_password_redacted_failures():
    name = "IIS:\\"
    container = "AppPools"
    settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "Sup3rS3cr3tP@ssW0rd",
        "processModel.identityType": "SpecificUser",
    }
    old_settings = {
        "processModel.userName": "Spongebob",
        "processModel.password": "0ldP@ssW0rd1!",
        "processModel.identityType": "SpecificUser",
    }
    current_settings = {
        "processModel.userName": "Administrator",
        "processModel.password": "0ldP@ssW0rd1!",
        "processModel.identityType": "SpecificUser",
    }

    new_settings = old_settings
    expected_ret = {
        "name": name,
        "changes": {
            "changes": {
                "processModel.userName": {"new": "Administrator", "old": "Spongebob"}
            },
            "failures": {
                "processModel.password": {
                    "new": "XXX-REDACTED-XXX",
                    "old": "XXX-REDACTED-XXX",
                }
            },
        },
        "comment": "Some settings failed to change.",
        "result": False,
    }
    with patch.dict(
        win_iis.__salt__,
        {
            "win_iis.get_container_setting": MagicMock(
                side_effect=[old_settings, current_settings, new_settings]
            ),
            "win_iis.set_container_setting": MagicMock(return_value=True),
        },
    ), patch.dict(win_iis.__opts__, {"test": False}):
        actual_ret = win_iis.container_setting(
            name=name, container=container, settings=settings
        )
    assert expected_ret == actual_ret



import pytest
import salt.modules.win_task as win_task
import salt.utils.platform
from tests.support.unit import TestCase, skipIf


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
@pytest.mark.destructive_test
class WinTaskTestCase(TestCase):
    """
    Test cases for salt.modules.win_task
    """

    def test_repeat_interval(self):
        task_name = "SaltTest1"
        try:
            ret = win_task.create_task(
                task_name,
                user_name="System",
                force=True,
                action_type="Execute",
                cmd="c:\\salt\\salt-call.bat",
                trigger_type="Daily",
                trigger_enabled=True,
                repeat_duration="30 minutes",
                repeat_interval="30 minutes",
            )
            self.assertTrue(ret)

            ret = win_task.info(task_name)
            self.assertEqual(ret["triggers"][0]["trigger_type"], "Daily")
        finally:
            ret = win_task.delete_task(task_name)
            self.assertTrue(ret)

    def test_repeat_interval_and_indefinitely(self):
        task_name = "SaltTest2"
        try:
            ret = win_task.create_task(
                task_name,
                user_name="System",
                force=True,
                action_type="Execute",
                cmd="c:\\salt\\salt-call.bat",
                trigger_type="Daily",
                trigger_enabled=True,
                repeat_duration="Indefinitely",
                repeat_interval="30 minutes",
            )
            self.assertTrue(ret)

            ret = win_task.info(task_name)
            self.assertEqual(ret["triggers"][0]["trigger_type"], "Daily")
        finally:
            ret = win_task.delete_task(task_name)
            self.assertTrue(ret)



"""
 Namecheap Management library of common functions used by
 all the namecheap execution modules

 Installation Prerequisites
 --------------------------

 - This module uses the following python libraries to communicate to
   the namecheap API:

        * ``requests``
        .. code-block:: bash

            pip install requests

"""


import logging
import xml.dom.minidom

import salt.loader

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

# Get logging started
log = logging.getLogger(__name__)

__salt__ = None


def __virtual__():
    if not HAS_REQUESTS:
        return (
            False,
            "Missing dependency: 'requests'. The namecheap utils module "
            "cannot be loaded. ",
        )
    global __salt__
    if not __salt__:
        __salt__ = salt.loader.minion_mods(__opts__)
    return True


def post_request(opts):
    namecheap_url = __salt__["config.option"]("namecheap.url")
    return _handle_request(requests.post(namecheap_url, data=opts, timeout=45))


def get_request(opts):
    namecheap_url = __salt__["config.option"]("namecheap.url")
    return _handle_request(requests.get(namecheap_url, params=opts, timeout=45))


def _handle_request(r):
    r.close()

    if r.status_code > 299:
        log.error(str(r))
        raise Exception(str(r))

    response_xml = xml.dom.minidom.parseString(r.text)
    apiresponse = response_xml.getElementsByTagName("ApiResponse")[0]

    if apiresponse.getAttribute("Status") == "ERROR":
        data = []
        errors = apiresponse.getElementsByTagName("Errors")[0]
        for e in errors.getElementsByTagName("Error"):
            data.append(e.firstChild.data)
        error = "".join(data)
        log.info(apiresponse)
        log.error(error)
        raise Exception(error)

    return response_xml


def xml_to_dict(xml):
    if xml.nodeType == xml.CDATA_SECTION_NODE:
        return xml.data
    result = atts_to_dict(xml)
    if len([n for n in xml.childNodes if n.nodeType != xml.TEXT_NODE]) == 0:
        if len(result) > 0:
            if xml.firstChild is not None and len(xml.firstChild.data) > 0:
                result["data"] = xml.firstChild.data
        elif xml.firstChild is not None and len(xml.firstChild.data) > 0:
            return xml.firstChild.data
        else:
            return None
    elif (
        xml.childNodes.length == 1
        and xml.childNodes[0].nodeType == xml.CDATA_SECTION_NODE
    ):
        return xml.childNodes[0].data
    else:
        for n in xml.childNodes:
            if n.nodeType == xml.CDATA_SECTION_NODE:

                if xml.tagName.lower() in result:
                    val = result[xml.tagName.lower()]
                    if not isinstance(val, list):
                        temp = [val]
                        val = temp
                    val.append(n.data)
                    result[xml.tagName.lower()] = val
                else:
                    result[xml.tagName.lower()] = n.data

            elif n.nodeType != xml.TEXT_NODE:

                if n.tagName.lower() in result:
                    val = result[n.tagName.lower()]

                    if not isinstance(val, list):
                        temp = [val]
                        val = temp
                    val.append(xml_to_dict(n))
                    result[n.tagName.lower()] = val
                else:
                    result[n.tagName.lower()] = xml_to_dict(n)
    return result


def atts_to_dict(xml):
    result = {}
    if xml.attributes is not None:
        for key, value in xml.attributes.items():
            result[key.lower()] = string_to_value(value)
    return result


def string_to_value(value):
    temp = value.lower()
    result = None
    if temp == "true":
        result = True
    elif temp == "false":
        result = False
    else:
        try:
            result = int(value)
        except ValueError:
            try:
                result = float(value)
            except ValueError:
                result = value

    return result


def get_opts(command):
    opts = {}
    opts["ApiUser"] = __salt__["config.option"]("namecheap.name")
    opts["UserName"] = __salt__["config.option"]("namecheap.user")
    opts["ApiKey"] = __salt__["config.option"]("namecheap.key")
    opts["ClientIp"] = __salt__["config.option"]("namecheap.client_ip")
    opts["Command"] = command
    return opts



import salt.modules.config as config
import salt.modules.zenoss as zenoss
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase


class ZenossTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.keystone
    """

    def setup_loader_modules(self):
        return {
            zenoss: {"__salt__": {"config.option": config.option}},
            config: {"__opts__": {}},
        }

    def test_zenoss_session(self):
        """
        test zenoss._session when using verify_ssl
        """
        zenoss_conf = {
            "zenoss": {
                "hostname": "https://test.zenoss.com",
                "username": "admin",
                "password": "test123",
            }
        }

        for verify in [True, False, None]:
            zenoss_conf["zenoss"]["verify_ssl"] = verify
            if verify is None:
                zenoss_conf["zenoss"].pop("verify_ssl")
                verify = True

            patch_opts = patch.dict(config.__opts__, zenoss_conf)
            mock_http = MagicMock(return_value=None)
            patch_http = patch("salt.utils.http.session", mock_http)

            with patch_http, patch_opts:
                zenoss._session()
                self.assertEqual(
                    mock_http.call_args_list,
                    [
                        call(
                            ca_bundle=None,
                            headers={"Content-type": "application/json; charset=utf-8"},
                            password="test123",
                            user="admin",
                            verify_ssl=verify,
                        )
                    ],
                )



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import os

import salt.modules.daemontools as daemontools
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class DaemontoolsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.daemontools
    """

    def setup_loader_modules(self):
        return {daemontools: {}}

    def test_start(self):
        """
        Test for Starts service via daemontools
        """
        mock = MagicMock(return_value=None)
        with patch.dict(daemontools.__salt__, {"file.remove": mock}):
            mock = MagicMock(return_value="")
            with patch.object(daemontools, "_service_path", mock):
                mock = MagicMock(return_value=False)
                with patch.dict(daemontools.__salt__, {"cmd.retcode": mock}):
                    self.assertTrue(daemontools.start("name"))

    def test_stop(self):
        """
        Test for Stops service via daemontools
        """
        mock = MagicMock(return_value=None)
        with patch.dict(daemontools.__salt__, {"file.touch": mock}):
            mock = MagicMock(return_value="")
            with patch.object(daemontools, "_service_path", mock):
                mock = MagicMock(return_value=False)
                with patch.dict(daemontools.__salt__, {"cmd.retcode": mock}):
                    self.assertTrue(daemontools.stop("name"))

    def test_term(self):
        """
        Test for Send a TERM to service via daemontools
        """
        mock = MagicMock(return_value="")
        with patch.object(daemontools, "_service_path", mock):
            mock = MagicMock(return_value=False)
            with patch.dict(daemontools.__salt__, {"cmd.retcode": mock}):
                self.assertTrue(daemontools.term("name"))

    def test_reload_(self):
        """
        Test for Wrapper for term()
        """
        mock = MagicMock(return_value=None)
        with patch.object(daemontools, "term", mock):
            self.assertEqual(daemontools.reload_("name"), None)

    def test_restart(self):
        """
        Test for Restart service via daemontools. This will stop/start service
        """
        mock = MagicMock(return_value=False)
        with patch.object(daemontools, "stop", mock):
            self.assertEqual(daemontools.restart("name"), "restart False")

    def test_full_restart(self):
        """
        Test for Calls daemontools.restart() function
        """
        mock = MagicMock(return_value=None)
        with patch.object(daemontools, "restart", mock):
            self.assertEqual(daemontools.restart("name"), None)

    def test_status(self):
        """
        Test for Return the status for a service via
        daemontools, return pid if running
        """
        with patch("re.search", MagicMock(return_value=1)):
            mock = MagicMock(return_value="")
            with patch.object(daemontools, "_service_path", mock):
                mock = MagicMock(return_value="name")
                with patch.dict(daemontools.__salt__, {"cmd.run_stdout": mock}):
                    self.assertEqual(daemontools.status("name"), "")

    def test_available(self):
        """
        Test for Returns ``True`` if the specified service
        is available, otherwise returns``False``.
        """
        mock = MagicMock(return_value=[])
        with patch.object(daemontools, "get_all", mock):
            self.assertFalse(daemontools.available("name"))

    def test_missing(self):
        """
        Test for The inverse of daemontools.available.
        """
        mock = MagicMock(return_value=[])
        with patch.object(daemontools, "get_all", mock):
            self.assertTrue(daemontools.missing("name"))

    def test_get_all(self):
        """
        Test for Return a list of all available services
        """
        self.assertRaises(CommandExecutionError, daemontools.get_all)

        with patch.object(daemontools, "SERVICE_DIR", "A"):
            mock = MagicMock(return_value="A")
            with patch.object(os, "listdir", mock):
                self.assertEqual(daemontools.get_all(), ["A"])



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os

import salt.modules.environ as environ
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class EnvironTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.environ
    """

    def setup_loader_modules(self):
        return {environ: {}}

    def test_setval(self):
        """
        Test for set a single salt process environment variable. Returns True
        on success.
        """
        mock = MagicMock(return_value=None)
        with patch.dict(os.environ, {}):
            self.assertEqual(environ.setval("key", False, True), None)

        mock = MagicMock(side_effect=Exception())
        with patch.dict(os.environ, {}):
            self.assertFalse(environ.setval("key", False, True))

        mock_environ = {}
        with patch.dict(os.environ, mock_environ):
            self.assertEqual(environ.setval("key", False), "")

        with patch.dict(os.environ, mock_environ):
            self.assertFalse(environ.setval("key", True))

    def test_set_val_permanent(self):
        with patch.dict(os.environ, {}), patch.dict(
            environ.__utils__,
            {"reg.set_value": MagicMock(), "reg.delete_value": MagicMock()},
        ), patch("salt.utils.platform.is_windows", return_value=True):

            environ.setval("key", "Test", permanent=True)
            environ.__utils__["reg.set_value"].assert_called_with(
                "HKCU", "Environment", "key", "Test"
            )

    def test_set_val_permanent_false_unsets(self):
        with patch.dict(os.environ, {}), patch.dict(
            environ.__utils__,
            {"reg.set_value": MagicMock(), "reg.delete_value": MagicMock()},
        ), patch("salt.utils.platform.is_windows", return_value=True):

            environ.setval("key", False, false_unsets=True, permanent=True)
            environ.__utils__["reg.set_value"].assert_not_called()
            environ.__utils__["reg.delete_value"].assert_called_with(
                "HKCU", "Environment", "key"
            )

    def test_set_val_permanent_hklm(self):
        with patch.dict(os.environ, {}), patch.dict(
            environ.__utils__,
            {"reg.set_value": MagicMock(), "reg.delete_value": MagicMock()},
        ), patch("salt.utils.platform.is_windows", return_value=True):

            key = r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment"
            environ.setval("key", "Test", permanent="HKLM")
            environ.__utils__["reg.set_value"].assert_called_with(
                "HKLM", key, "key", "Test"
            )

    def test_setenv(self):
        """
        Set multiple salt process environment variables from a dict.
        Returns a dict.
        """
        mock_environ = {"KEY": "value"}
        with patch.dict(os.environ, mock_environ):
            self.assertFalse(environ.setenv("environ"))

        with patch.dict(os.environ, mock_environ):
            self.assertFalse(environ.setenv({"A": True}, False, True, False))

        with patch.dict(os.environ, mock_environ):
            mock_setval = MagicMock(return_value=None)
            with patch.object(environ, "setval", mock_setval):
                self.assertEqual(environ.setenv({}, False, True, False)["KEY"], None)

    def test_get(self):
        """
        Get a single salt process environment variable.
        """
        self.assertFalse(environ.get(True))

        self.assertEqual(environ.get("key"), "")

    def test_has_value(self):
        """
        Determine whether the key exists in the current salt process
        environment dictionary. Optionally compare the current value
        of the environment against the supplied value string.
        """
        mock_environ = {}
        with patch.dict(os.environ, mock_environ):
            self.assertFalse(environ.has_value(True))

            os.environ["salty"] = "yes"
            self.assertTrue(environ.has_value("salty", "yes"))

            os.environ["too_salty"] = "no"
            self.assertFalse(environ.has_value("too_salty", "yes"))

            self.assertFalse(environ.has_value("key", "value"))

            os.environ["key"] = "value"
            self.assertTrue(environ.has_value("key"))

    def test_item(self):
        """
        Get one or more salt process environment variables.
        Returns a dict.
        """
        self.assertEqual(environ.item(None), {})

    def test_items(self):
        """
        Return a dict of the entire environment set for the salt process
        """
        self.assertNotEqual(list(environ.items()), [])



"""
Manage Linux kernel packages on YUM-based systems
"""

import functools
import logging

try:
    from salt.utils.versions import LooseVersion as _LooseVersion
    from salt.exceptions import CommandExecutionError
    import salt.utils.data
    import salt.utils.functools
    import salt.utils.systemd
    import salt.modules.yumpkg

    __IMPORT_ERROR = None
except ImportError as exc:
    __IMPORT_ERROR = exc.__str__()

log = logging.getLogger(__name__)

__virtualname__ = "kernelpkg"

# Import functions from yumpkg
# pylint: disable=invalid-name, protected-access
_yum = salt.utils.functools.namespaced_function(salt.modules.yumpkg._yum, globals())
# pylint: enable=invalid-name, protected-access


def __virtual__():
    """
    Load this module on RedHat-based systems only
    """

    if __IMPORT_ERROR:
        return (False, __IMPORT_ERROR)

    if __grains__.get("os_family", "") == "RedHat":
        return __virtualname__
    elif __grains__.get("os", "").lower() in (
        "amazon",
        "xcp",
        "xenserver",
        "virtuozzolinux",
    ):
        return __virtualname__

    return (False, "Module kernelpkg_linux_yum: no YUM based system detected")


def active():
    """
    Return the version of the running kernel.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.active
    """
    if "pkg.normalize_name" in __salt__:
        return __salt__["pkg.normalize_name"](__grains__["kernelrelease"])

    return __grains__["kernelrelease"]


def list_installed():
    """
    Return a list of all installed kernels.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.list_installed
    """
    result = __salt__["pkg.version"](_package_name(), versions_as_list=True)
    if result is None:
        return []

    return sorted(result, key=functools.cmp_to_key(_cmp_version))


def latest_available():
    """
    Return the version of the latest kernel from the package repositories.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.latest_available
    """
    result = __salt__["pkg.latest_version"](_package_name())
    if result == "":
        result = latest_installed()
    return result


def latest_installed():
    """
    Return the version of the latest installed kernel.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.latest_installed

    .. note::
        This function may not return the same value as
        :py:func:`~salt.modules.kernelpkg_linux_yum.active` if a new kernel
        has been installed and the system has not yet been rebooted.
        The :py:func:`~salt.modules.kernelpkg_linux_yum.needs_reboot` function
        exists to detect this condition.
    """
    pkgs = list_installed()
    if pkgs:
        return pkgs[-1]

    return None


def needs_reboot():
    """
    Detect if a new kernel version has been installed but is not running.
    Returns True if a new kernel is installed, False otherwise.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.needs_reboot
    """
    return _LooseVersion(active()) < _LooseVersion(latest_installed())


def upgrade(reboot=False, at_time=None):
    """
    Upgrade the kernel and optionally reboot the system.

    reboot : False
        Request a reboot if a new kernel is available.

    at_time : immediate
        Schedule the reboot at some point in the future. This argument
        is ignored if ``reboot=False``. See
        :py:func:`~salt.modules.system.reboot` for more details
        on this argument.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.upgrade
        salt '*' kernelpkg.upgrade reboot=True at_time=1

    .. note::
        An immediate reboot often shuts down the system before the minion has a
        chance to return, resulting in errors. A minimal delay (1 minute) is
        useful to ensure the result is delivered to the master.
    """
    result = __salt__["pkg.upgrade"](name=_package_name())
    _needs_reboot = needs_reboot()

    ret = {
        "upgrades": result,
        "active": active(),
        "latest_installed": latest_installed(),
        "reboot_requested": reboot,
        "reboot_required": _needs_reboot,
    }

    if reboot and _needs_reboot:
        log.warning("Rebooting system due to kernel upgrade")
        __salt__["system.reboot"](at_time=at_time)

    return ret


def upgrade_available():
    """
    Detect if a new kernel version is available in the repositories.
    Returns True if a new kernel is available, False otherwise.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.upgrade_available
    """
    return _LooseVersion(latest_available()) > _LooseVersion(latest_installed())


def remove(release):
    """
    Remove a specific version of the kernel.

    release
        The release number of an installed kernel. This must be the entire release
        number as returned by :py:func:`~salt.modules.kernelpkg_linux_yum.list_installed`,
        not the package name.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.remove 3.10.0-327.el7
    """
    if release not in list_installed():
        raise CommandExecutionError(
            "Kernel release '{}' is not installed".format(release)
        )

    if release == active():
        raise CommandExecutionError("Active kernel cannot be removed")

    target = "{}-{}".format(_package_name(), release)
    log.info("Removing kernel package %s", target)
    old = __salt__["pkg.list_pkgs"]()

    # Build the command string
    cmd = []
    if salt.utils.systemd.has_scope(__context__) and __salt__["config.get"](
        "systemd.scope", True
    ):
        cmd.extend(["systemd-run", "--scope"])
    cmd.extend([_yum(), "-y", "remove", target])  # pylint: disable=not-callable

    # Execute the command
    out = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    # Look for the changes in installed packages
    __context__.pop("pkg.list_pkgs", None)
    new = __salt__["pkg.list_pkgs"]()
    ret = salt.utils.data.compare_dicts(old, new)

    # Look for command execution errors
    if out["retcode"] != 0:
        raise CommandExecutionError(
            "Error occurred removing package(s)",
            info={"errors": [out["stderr"]], "changes": ret},
        )

    return {"removed": [target]}


def cleanup(keep_latest=True):
    """
    Remove all unused kernel packages from the system.

    keep_latest : True
        In the event that the active kernel is not the latest one installed, setting this to True
        will retain the latest kernel package, in addition to the active one. If False, all kernel
        packages other than the active one will be removed.

    CLI Example:

    .. code-block:: bash

        salt '*' kernelpkg.cleanup
    """
    removed = []

    # Loop over all installed kernel packages
    for kernel in list_installed():

        # Keep the active kernel package
        if kernel == active():
            continue

        # Optionally keep the latest kernel package
        if keep_latest and kernel == latest_installed():
            continue

        # Remove the kernel package
        removed.extend(remove(kernel)["removed"])

    return {"removed": removed}


def _package_name():
    """
    Return static string for the package name
    """
    return "kernel"


def _cmp_version(item1, item2):
    """
    Compare function for package version sorting
    """
    vers1 = _LooseVersion(item1)
    vers2 = _LooseVersion(item2)

    if vers1 < vers2:
        return -1
    if vers1 > vers2:
        return 1
    return 0



"""
ACME / Let's Encrypt certificate management state
=================================================

.. versionadded:: 2016.3

See also the module documentation

.. code-block:: yaml

    reload-gitlab:
      cmd.run:
        - name: gitlab-ctl hup

    dev.example.com:
      acme.cert:
        - aliases:
          - gitlab.example.com
        - email: acmemaster@example.com
        - webroot: /opt/gitlab/embedded/service/gitlab-rails/public
        - renew: 14
        - fire_event: acme/dev.example.com
        - onchanges_in:
          - cmd: reload-gitlab

"""

import logging

import salt.utils.dictdiffer

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work when the ACME module agrees
    """
    if "acme.cert" in __salt__:
        return True
    return (False, "acme module could not be loaded")


def cert(
    name,
    aliases=None,
    email=None,
    webroot=None,
    test_cert=False,
    renew=None,
    keysize=None,
    server=None,
    owner="root",
    group="root",
    mode="0640",
    certname=None,
    preferred_challenges=None,
    tls_sni_01_port=None,
    tls_sni_01_address=None,
    http_01_port=None,
    http_01_address=None,
    dns_plugin=None,
    dns_plugin_credentials=None,
):
    """
    Obtain/renew a certificate from an ACME CA, probably Let's Encrypt.

    :param name: Common Name of the certificate (DNS name of certificate)
    :param aliases: subjectAltNames (Additional DNS names on certificate)
    :param email: e-mail address for interaction with ACME provider
    :param webroot: True or a full path to webroot. Otherwise use standalone mode
    :param test_cert: Request a certificate from the Happy Hacker Fake CA (mutually exclusive with 'server')
    :param renew: True/'force' to force a renewal, or a window of renewal before expiry in days
    :param keysize: RSA key bits
    :param server: API endpoint to talk to
    :param owner: owner of the private key file
    :param group: group of the private key file
    :param mode: mode of the private key file
    :param certname: Name of the certificate to save
    :param preferred_challenges: A sorted, comma delimited list of the preferred
                                 challenge to use during authorization with the
                                 most preferred challenge listed first.
    :param tls_sni_01_port: Port used during tls-sni-01 challenge. This only affects
                            the port Certbot listens on. A conforming ACME server
                            will still attempt to connect on port 443.
    :param tls_sni_01_address: The address the server listens to during tls-sni-01
                               challenge.
    :param http_01_port: Port used in the http-01 challenge. This only affects
                         the port Certbot listens on. A conforming ACME server
                         will still attempt to connect on port 80.
    :param https_01_address: The address the server listens to during http-01 challenge.
    :param dns_plugin: Name of a DNS plugin to use (currently only 'cloudflare')
    :param dns_plugin_credentials: Path to the credentials file if required by the specified DNS plugin
    """

    if certname is None:
        certname = name

    ret = {"name": certname, "result": "changeme", "comment": [], "changes": {}}
    action = None

    current_certificate = {}
    new_certificate = {}
    if not __salt__["acme.has"](certname):
        action = "obtain"
    elif __salt__["acme.needs_renewal"](certname, renew):
        action = "renew"
        current_certificate = __salt__["acme.info"](certname)
    else:
        ret["result"] = True
        ret["comment"].append(
            "Certificate {} exists and does not need renewal.".format(certname)
        )

    if action:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"].append(
                "Certificate {} would have been {}ed.".format(certname, action)
            )
            ret["changes"] = {"old": "current certificate", "new": "new certificate"}
        else:
            res = __salt__["acme.cert"](
                name,
                aliases=aliases,
                email=email,
                webroot=webroot,
                certname=certname,
                test_cert=test_cert,
                renew=renew,
                keysize=keysize,
                server=server,
                owner=owner,
                group=group,
                mode=mode,
                preferred_challenges=preferred_challenges,
                tls_sni_01_port=tls_sni_01_port,
                tls_sni_01_address=tls_sni_01_address,
                http_01_port=http_01_port,
                http_01_address=http_01_address,
                dns_plugin=dns_plugin,
                dns_plugin_credentials=dns_plugin_credentials,
            )
            ret["result"] = res["result"]
            ret["comment"].append(res["comment"])
            if ret["result"]:
                new_certificate = __salt__["acme.info"](certname)
            ret["changes"] = salt.utils.dictdiffer.deep_diff(
                current_certificate, new_certificate
            )
    return ret



"""
Pass Renderer for Salt
======================

pass_ is an encrypted on-disk password store.

.. _pass: https://www.passwordstore.org/

.. versionadded:: 2017.7.0

Setup
-----

*Note*: ``<user>`` needs to be replaced with the user salt-master will be
running as.

Have private gpg loaded into ``user``'s gpg keyring

.. code-block:: yaml

    load_private_gpg_key:
      cmd.run:
        - name: gpg --import <location_of_private_gpg_key>
        - unless: gpg --list-keys '<gpg_name>'

Said private key's public key should have been used when encrypting pass entries
that are of interest for pillar data.

Fetch and keep local pass git repo up-to-date

.. code-block:: yaml

        update_pass:
          git.latest:
            - force_reset: True
            - name: <git_repo>
            - target: /<user>/.password-store
            - identity: <location_of_ssh_private_key>
            - require:
              - cmd: load_private_gpg_key

Install pass binary

.. code-block:: yaml

        pass:
          pkg.installed
"""


import logging
import os
from os.path import expanduser
from subprocess import PIPE, Popen

import salt.utils.path
from salt.exceptions import SaltRenderError

log = logging.getLogger(__name__)


def _get_pass_exec():
    """
    Return the pass executable or raise an error
    """
    pass_exec = salt.utils.path.which("pass")
    if pass_exec:
        return pass_exec
    else:
        raise SaltRenderError("pass unavailable")


def _fetch_secret(pass_path):
    """
    Fetch secret from pass based on pass_path. If there is
    any error, return back the original pass_path value
    """
    cmd = "pass show {}".format(pass_path.strip())
    log.debug("Fetching secret: %s", cmd)

    proc = Popen(cmd.split(" "), stdout=PIPE, stderr=PIPE)
    pass_data, pass_error = proc.communicate()

    # The version of pass used during development sent output to
    # stdout instead of stderr even though its returncode was non zero.
    if proc.returncode or not pass_data:
        log.warning("Could not fetch secret: %s %s", pass_data, pass_error)
        pass_data = pass_path
    return pass_data.strip()


def _decrypt_object(obj):
    """
    Recursively try to find a pass path (string) that can be handed off to pass
    """
    if isinstance(obj, str):
        return _fetch_secret(obj)
    elif isinstance(obj, dict):
        for pass_key, pass_path in obj.items():
            obj[pass_key] = _decrypt_object(pass_path)
    elif isinstance(obj, list):
        for pass_key, pass_path in enumerate(obj):
            obj[pass_key] = _decrypt_object(pass_path)
    return obj


def render(pass_info, saltenv="base", sls="", argline="", **kwargs):
    """
    Fetch secret from pass based on pass_path
    """
    _get_pass_exec()

    # Make sure environment variable HOME is set, since Pass looks for the
    # password-store under ~/.password-store.
    os.environ["HOME"] = expanduser("~")
    return _decrypt_object(pass_info)



"""
Return salt data via pushover (http://www.pushover.net)

.. versionadded:: 2016.3.0

The following fields can be set in the minion conf file::

    pushover.user (required)
    pushover.token (required)
    pushover.title (optional)
    pushover.device (optional)
    pushover.priority (optional)
    pushover.expire (optional)
    pushover.retry (optional)
    pushover.profile (optional)

.. note::
    The ``user`` here is your **user key**, *not* the email address you use to
    login to pushover.net.

Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location::

    alternative.pushover.user
    alternative.pushover.token
    alternative.pushover.title
    alternative.pushover.device
    alternative.pushover.priority
    alternative.pushover.expire
    alternative.pushover.retry

PushOver settings may also be configured as::

    pushover:
        user: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        token: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        title: Salt Returner
        device: phone
        priority: -1
        expire: 3600
        retry: 5

    alternative.pushover:
        user: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        token: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        title: Salt Returner
        device: phone
        priority: 1
        expire: 4800
        retry: 2

    pushover_profile:
        pushover.token: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    pushover:
        user: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        profile: pushover_profile

    alternative.pushover:
        user: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        profile: pushover_profile

  To use the PushOver returner, append '--return pushover' to the salt command. ex:

  .. code-block:: bash

    salt '*' test.ping --return pushover

  To use the alternative configuration, append '--return_config alternative' to the salt command. ex:

    salt '*' test.ping --return pushover --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return pushover --return_kwargs '{"title": "Salt is awesome!"}'

"""

import logging
import pprint
import urllib.parse

import salt.returners
import salt.utils.pushover
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

__virtualname__ = "pushover"


def _get_options(ret=None):
    """
    Get the pushover options from salt.
    """

    defaults = {"priority": "0"}

    attrs = {
        "pushover_profile": "profile",
        "user": "user",
        "device": "device",
        "token": "token",
        "priority": "priority",
        "title": "title",
        "api_version": "api_version",
        "expire": "expire",
        "retry": "retry",
        "sound": "sound",
    }

    profile_attr = "pushover_profile"

    profile_attrs = {
        "user": "user",
        "device": "device",
        "token": "token",
        "priority": "priority",
        "title": "title",
        "api_version": "api_version",
        "expire": "expire",
        "retry": "retry",
        "sound": "sound",
    }

    _options = salt.returners.get_returner_options(
        __virtualname__,
        ret,
        attrs,
        profile_attr=profile_attr,
        profile_attrs=profile_attrs,
        __salt__=__salt__,
        __opts__=__opts__,
        defaults=defaults,
    )
    return _options


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """
    return __virtualname__


def _post_message(
    user,
    device,
    message,
    title,
    priority,
    expire,
    retry,
    sound,
    api_version=1,
    token=None,
):
    """
    Send a message to a Pushover user or group.
    :param user:        The user or group to send to, must be key of user or group not email address.
    :param message:     The message to send to the PushOver user or group.
    :param title:       Specify who the message is from.
    :param priority     The priority of the message, defaults to 0.
    :param api_version: The PushOver API version, if not specified in the configuration.
    :param notify:      Whether to notify the room, default: False.
    :param token:       The PushOver token, if not specified in the configuration.
    :return:            Boolean if message was sent successfully.
    """

    user_validate = salt.utils.pushover.validate_user(user, device, token)
    if not user_validate["result"]:
        return user_validate

    parameters = dict()
    parameters["user"] = user
    parameters["device"] = device
    parameters["token"] = token
    parameters["title"] = title
    parameters["priority"] = priority
    parameters["expire"] = expire
    parameters["retry"] = retry
    parameters["message"] = message

    if sound:
        sound_validate = salt.utils.pushover.validate_sound(sound, token)
        if sound_validate["res"]:
            parameters["sound"] = sound

    result = salt.utils.pushover.query(
        function="message",
        method="POST",
        header_dict={"Content-Type": "application/x-www-form-urlencoded"},
        data=urllib.parse.urlencode(parameters),
        opts=__opts__,
    )

    return result


def returner(ret):
    """
    Send an PushOver message with the data
    """

    _options = _get_options(ret)

    user = _options.get("user")
    device = _options.get("device")
    token = _options.get("token")
    title = _options.get("title")
    priority = _options.get("priority")
    expire = _options.get("expire")
    retry = _options.get("retry")
    sound = _options.get("sound")

    if not token:
        raise SaltInvocationError("Pushover token is unavailable.")

    if not user:
        raise SaltInvocationError("Pushover user key is unavailable.")

    if priority and priority == 2:
        if not expire and not retry:
            raise SaltInvocationError(
                "Priority 2 requires pushover.expire and pushover.retry options."
            )

    message = "id: {}\r\nfunction: {}\r\nfunction args: {}\r\njid: {}\r\nreturn: {}\r\n".format(
        ret.get("id"),
        ret.get("fun"),
        ret.get("fun_args"),
        ret.get("jid"),
        pprint.pformat(ret.get("return")),
    )

    result = _post_message(
        user=user,
        device=device,
        message=message,
        title=title,
        priority=priority,
        expire=expire,
        retry=retry,
        sound=sound,
        token=token,
    )

    log.debug("pushover result %s", result)
    if not result["res"]:
        log.info("Error: %s", result["message"])
    return



"""
Integration tests for the rabbitmq_cluster states
"""

import logging

import pytest
import salt.modules.rabbitmq as rabbitmq
import salt.states.rabbitmq_cluster as rabbitmq_cluster

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_freebsd(reason="No Docker on FreeBSD available"),
    pytest.mark.skip_if_binaries_missing(
        "docker", "dockerd", reason="Docker not installed"
    ),
]


@pytest.fixture
def configure_loader_modules(docker_cmd_run_all_wrapper):
    return {
        rabbitmq_cluster: {
            "__salt__": {
                "rabbitmq.cluster_status": rabbitmq.cluster_status,
                "rabbitmq.join_cluster": rabbitmq.join_cluster,
            },
            "__opts__": {"test": False},
            "_utils__": {},
        },
        rabbitmq: {
            "__salt__": {"cmd.run_all": docker_cmd_run_all_wrapper},
            "__opts__": {},
            "_utils__": {},
        },
    }


@pytest.mark.slow_test
def test_joined(rabbitmq_container):
    """
    Test rabbitmq_cluster.joined
    """

    hostname = rabbitmq_container.container.attrs["Config"]["Hostname"]
    ret = rabbitmq_cluster.joined("name", host=hostname)
    expected = {
        "name": "name",
        "result": True,
        "comment": "Already in cluster",
        "changes": {},
    }
    assert ret == expected
